# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from more_itertools import first

from orchestrator.db import ProcessTable
from orchestrator.forms import generate_form
from orchestrator.workflow import ProcessStat


def show_process(process: ProcessTable, pStat: ProcessStat) -> dict:
    subscription = first(process.subscriptions, None)
    if subscription:
        product_id = subscription.product_id
        customer_id = subscription.customer_id
    else:
        product_id = None
        customer_id = None

    steps = [
        {
            "name": step.name,
            "executed": step.executed_at.timestamp(),
            "status": step.status,
            "state": step.state,
            "created_by": step.created_by,
            "stepid": step.stepid,
        }
        for step in process.steps
    ]

    form = None
    if pStat.log:
        form = pStat.log[0].form
        pstat_steps = list(map(lambda step: {"name": step.name, "status": "pending"}, pStat.log))
        steps += pstat_steps

    current_state = pStat.state.unwrap() if pStat.state else None
    generated_form = generate_form(form, current_state, []) if form and current_state else None

    return {
        "id": process.pid,
        "pid": process.pid,  # list and single get differentiate with this value and the above.
        "workflow": process.workflow,
        "workflow_name": process.workflow,
        "product": product_id,
        "customer": customer_id,
        "assignee": process.assignee,
        "status": process.last_status,
        "last_status": process.last_status,  # list and single get differentiate with this value and the above.
        "failed_reason": process.failed_reason,
        "traceback": process.traceback,
        "step": process.last_step,
        "steps": steps,
        "created_by": process.created_by,
        "started": process.started_at.timestamp(),
        "last_modified": process.last_modified_at.timestamp(),
        "subscriptions": [
            # explicit conversion using excluded_keys to prevent eager loaded subscriptions (when loaded for form domain models)
            # to cause circular reference errors
            s.subscription.__json__(excluded_keys={"instances", "customer_descriptions", "processes", "product"})
            for s in process.process_subscriptions
        ],
        "is_task": process.is_task,
        "form": generated_form,
        "current_state": current_state,
    }



# Copyright 2019-2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Tuple

from redis import Redis
from redis.asyncio import Redis as AIORedis
from redis.asyncio.lock import Lock
from redis.exceptions import LockError
from redis.lock import Lock as SyncLock
from structlog import get_logger

logger = get_logger(__name__)


class RedisDistLockManager:
    """Create Distributed Locks in Redis.

    https://redis.io/topics/distlock
    """

    namespace = "orchestrator:distlock"

    def __init__(self, redis_address: Tuple[str, int]):
        self.redis_conn: Optional[AIORedis] = None
        self.redis_address = redis_address

    async def connect_redis(self) -> None:
        self.redis_conn = AIORedis(host=self.redis_address[0], port=self.redis_address[1])

    async def disconnect_redis(self) -> None:
        if self.redis_conn:
            await self.redis_conn.close()

    async def get_lock(self, resource: str, expiration_seconds: int) -> Optional[Lock]:
        if not self.redis_conn:
            return None

        key = f"{self.namespace}:{resource}"
        try:
            lock: Lock = Lock(
                redis=self.redis_conn,
                name=key,
                timeout=float(expiration_seconds),
                blocking=False,
                thread_local=False,
            )
            if await lock.acquire():
                return lock
            else:
                # normal behavior (lock acquired by something else)
                return None
        except LockError:
            # Unexpected behavior, possibly a problem with Redis
            logger.Exception("Could not acquire lock for resource", resource=key)
            return None

    async def release_lock(self, lock: Lock) -> None:
        if not self.redis_conn:
            return None

        try:
            await lock.release()
        except LockError:
            logger.Exception("Could not release lock for resource", resource=lock.name)

    # https://github.com/aio-libs/aioredis-py/issues/1273
    def release_sync(self, lock: Lock) -> None:
        redis_conn: Optional[Redis] = None
        try:
            redis_conn = Redis(host=self.redis_address[0], port=self.redis_address[1])
            sync_lock: SyncLock = SyncLock(
                redis=redis_conn,
                name=lock.name,
                timeout=lock.timeout,
                blocking=False,
                thread_local=False,
            )
            sync_lock.local = lock.local
            sync_lock.release()
        except LockError:
            logger.Exception("Could not release lock for resource", resource=lock.name)
        finally:
            if redis_conn:
                redis_conn.close()


__all__ = [
    "RedisDistLockManager",
    "Lock",
]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import getenv
from typing import Any, Dict, Optional, Tuple
from uuid import UUID

from redis import Redis
from structlog import get_logger

from orchestrator.services.subscriptions import _generate_etag
from orchestrator.settings import app_settings
from orchestrator.utils.json import json_dumps, json_loads

logger = get_logger(__name__)

cache = Redis(host=app_settings.CACHE_HOST, port=app_settings.CACHE_PORT)

ONE_WEEK = 3600 * 24 * 7


def caching_models_enabled() -> bool:
    return getenv("AIOCACHE_DISABLE", 0) == 0 and app_settings.CACHE_DOMAIN_MODELS


def to_redis(subscription: Dict[str, Any]) -> None:
    if caching_models_enabled():
        logger.info("Setting cache for subscription.", subscription=subscription["subscription_id"])
        etag = _generate_etag(subscription)
        cache.set(f"domain:{subscription['subscription_id']}", json_dumps(subscription), ex=ONE_WEEK)
        cache.set(f"domain:etag:{subscription['subscription_id']}", etag, ex=ONE_WEEK)
    else:
        logger.warning("Caching disabled, not caching subscription", subscription=subscription["subscription_id"])


def from_redis(subscription_id: UUID) -> Optional[Tuple[Any, str]]:
    if caching_models_enabled():
        logger.info("Retrieving subscription from cache", subscription=subscription_id)
        obj = cache.get(f"domain:{subscription_id}")
        etag = cache.get(f"domain:etag:{subscription_id}")
        if obj and etag:
            return json_loads(obj), etag.decode("utf-8")  # type: ignore
        else:
            return None
    else:
        logger.warning("Caching disabled, not loading subscription", subscription=subscription_id)
        return None


def delete_from_redis(subscription_id: UUID) -> None:
    if caching_models_enabled():
        logger.info("Deleting subscription object from cache", subscription_id=subscription_id)
        cache.delete(f"domain:{subscription_id}")
        cache.delete(f"domain:etag:{subscription_id}")
    else:
        logger.warning("Caching disabled, not deleting subscription", subscription=subscription_id)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orchestrator.schemas.engine_settings import EngineSettingsBaseSchema, EngineSettingsSchema, GlobalStatusEnum
from orchestrator.schemas.fixed_input import FixedInputConfigurationSchema, FixedInputSchema
from orchestrator.schemas.problem_detail import ProblemDetailSchema
from orchestrator.schemas.process import (
    ProcessBaseSchema,
    ProcessIdSchema,
    ProcessListItemSchema,
    ProcessResumeAllSchema,
    ProcessSchema,
    ProcessSubscriptionBaseSchema,
    ProcessSubscriptionSchema,
)
from orchestrator.schemas.product import ProductBaseSchema, ProductCRUDSchema, ProductSchema
from orchestrator.schemas.product_block import ProductBlockBaseSchema, ProductBlockEnrichedSchema
from orchestrator.schemas.resource_type import ResourceTypeBaseSchema, ResourceTypeSchema
from orchestrator.schemas.subscription import SubscriptionDomainModelSchema, SubscriptionIdSchema, SubscriptionSchema
from orchestrator.schemas.subscription_descriptions import (
    SubscriptionDescriptionBaseSchema,
    SubscriptionDescriptionSchema,
)
from orchestrator.schemas.workflow import SubscriptionWorkflowListsSchema, WorkflowSchema, WorkflowWithProductTagsSchema

__all__ = (
    "EngineSettingsSchema",
    "EngineSettingsBaseSchema",
    "FixedInputConfigurationSchema",
    "GlobalStatusEnum",
    "ProblemDetailSchema",
    "FixedInputSchema",
    "ProductBlockEnrichedSchema",
    "ProductBlockBaseSchema",
    "ProductCRUDSchema",
    "ProductBaseSchema",
    "ProductSchema",
    "ProcessSubscriptionSchema",
    "ProcessResumeAllSchema",
    "ProcessBaseSchema",
    "ProcessSchema",
    "ProcessIdSchema",
    "ProcessSubscriptionBaseSchema",
    "ProcessListItemSchema",
    "SubscriptionDescriptionBaseSchema",
    "SubscriptionDescriptionSchema",
    "SubscriptionSchema",
    "SubscriptionDomainModelSchema",
    "SubscriptionWorkflowListsSchema",
    "SubscriptionIdSchema",
    "ResourceTypeSchema",
    "ResourceTypeBaseSchema",
    "WorkflowSchema",
    "WorkflowWithProductTagsSchema",
)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ipaddress import IPv4Address, IPv6Address
from uuid import UUID

from structlog import get_logger
from test_orchestrator.products.test_product import TestProductInactive

from orchestrator.forms import FormPage
from orchestrator.forms.validators import OrganisationId
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, done, step, workflow
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form

logger = get_logger(__name__)


def initial_input_form_generator(product_name: str, product: UUIDstr) -> FormGenerator:
    class CreateTestProductForm(FormPage):
        class Config:
            title = product_name

        organisation: OrganisationId

        an_int: int
        a_str: str
        a_bool: bool
        an_uuid: UUIDstr
        an_ipv4: IPv4Address
        an_ipv6: IPv6Address

    user_input = yield CreateTestProductForm

    return user_input.dict()


@step("Construct Subscription model")
def construct_subscription_model(
    product: UUIDstr,
    organisation: UUIDstr,
    an_int: int,
    a_str: str,
    a_bool: bool,
    an_uuid: UUIDstr,
    an_ipv4: IPv4Address,
    an_ipv6: IPv6Address,
) -> State:
    test_product = TestProductInactive.from_product_id(
        product_id=product,
        customer_id=organisation,
        status=SubscriptionLifecycle.INITIAL,
    )
    test_product.testproduct.an_int = an_int
    test_product.testproduct.a_str = a_str
    test_product.testproduct.a_bool = a_bool
    test_product.testproduct.an_uuid = UUID(an_uuid)
    test_product.testproduct.an_ipv4 = an_ipv4
    test_product.testproduct.an_ipv6 = an_ipv6
    test_product.save()
    return {
        "subscription": test_product,
        "subscription_id": test_product.subscription_id,
        "subscription_description": test_product.description,
    }


@workflow("Create a test product", initial_input_form=wrap_create_initial_input_form(initial_input_form_generator))
def create_test_product() -> StepList:
    return begin >> construct_subscription_model >> store_process_subscription(Target.CREATE) >> done



import pytest

from orchestrator.forms import FormNotCompleteError, FormPage, FormValidationError, generate_form, post_process
from orchestrator.types import strEnum
from orchestrator.utils.json import json_dumps, json_loads


class TestChoices(strEnum):
    A = "a"
    B = "b"


class TestForm(FormPage):
    generic_select: TestChoices


def test_post_process_yield():
    def input_form(state):

        user_input = yield TestForm
        return {**user_input.dict(), "extra": 234}

    validated_data = post_process(input_form, {"previous": True}, [{"generic_select": "a"}])

    expected = {"generic_select": "a", "extra": 234}
    assert expected == json_loads(json_dumps(validated_data))


def test_post_process_extra_data():
    def input_form(state):
        user_input = yield TestForm
        return {**user_input.dict(), "extra": 234}

    with pytest.raises(FormValidationError) as e:
        post_process(input_form, {"previous": True}, [{"generic_select": "a", "extra_data": False}])

    assert (
        str(e.value)
        == "1 validation error for TestForm\nextra_data\n  extra fields not permitted (type=value_error.extra)"
    )


def test_post_process_validation_errors():
    def input_form(state):
        user_input = yield TestForm
        return user_input.dict()

    with pytest.raises(FormValidationError) as e:
        post_process(input_form, {}, [{"generic_select": 1, "extra_data": False}])

    assert (
        str(e.value)
        == "2 validation errors for TestForm\ngeneric_select\n  value is not a valid enumeration member; permitted: 'a', 'b' (type=type_error.enum; enum_values=[<TestChoices.A: 'a'>, <TestChoices.B: 'b'>])\nextra_data\n  extra fields not permitted (type=value_error.extra)"
    )

    with pytest.raises(FormValidationError) as e:
        post_process(input_form, {}, [{"generic_select": 1}])

    assert (
        str(e.value)
        == "1 validation error for TestForm\ngeneric_select\n  value is not a valid enumeration member; permitted: 'a', 'b' (type=type_error.enum; enum_values=[<TestChoices.A: 'a'>, <TestChoices.B: 'b'>])"
    )


def test_post_process_wizard():
    # Return if there is no form
    assert post_process(None, {}, []) == {}
    assert post_process([], {}, []) == {}

    def input_form(state):
        class TestForm1(FormPage):
            generic_select1: TestChoices

            class Config:
                title = "Some title"

        class TestForm2(FormPage):
            generic_select2: TestChoices

        class TestForm3(FormPage):
            generic_select3: TestChoices

        user_input_1 = yield TestForm1

        if user_input_1.generic_select1 == TestChoices.A:
            user_input_2 = yield TestForm2
        else:
            user_input_2 = yield TestForm3

        return {**user_input_1.dict(), **user_input_2.dict()}

    # Submit 1
    with pytest.raises(FormNotCompleteError) as error_info:
        post_process(input_form, {"previous": True}, [])

    assert error_info.value.form == {
        "title": "Some title",
        "type": "object",
        "additionalProperties": False,
        "definitions": {
            "TestChoices": {
                "description": "An enumeration.",
                "enum": ["a", "b"],
                "title": "TestChoices",
                "type": "string",
            }
        },
        "properties": {"generic_select1": {"$ref": "#/definitions/TestChoices"}},
        "required": ["generic_select1"],
    }

    # Submit 2
    with pytest.raises(FormNotCompleteError) as error_info:
        post_process(input_form, {"previous": True}, [{"generic_select1": "b"}])

    assert error_info.value.form == {
        "title": "unknown",
        "type": "object",
        "additionalProperties": False,
        "definitions": {
            "TestChoices": {
                "description": "An enumeration.",
                "enum": ["a", "b"],
                "title": "TestChoices",
                "type": "string",
            }
        },
        "properties": {"generic_select3": {"$ref": "#/definitions/TestChoices"}},
        "required": ["generic_select3"],
    }

    # Submit complete
    validated_data = post_process(input_form, {"previous": True}, [{"generic_select1": "b"}, {"generic_select3": "a"}])

    expected = {"generic_select1": "b", "generic_select3": "a"}
    assert expected == json_loads(json_dumps(validated_data))

    # Submit overcomplete
    validated_data = post_process(
        input_form, {"previous": True}, [{"generic_select1": "b"}, {"generic_select3": "a"}, {"to_much": True}]
    )

    expected = {"generic_select1": "b", "generic_select3": "a"}
    assert expected == json_loads(json_dumps(validated_data))


def test_generate_form():
    def input_form(state):
        class TestForm1(FormPage):
            generic_select1: TestChoices

            class Config:
                title = "Some title"

        class TestForm2(FormPage):
            generic_select2: TestChoices

        class TestForm3(FormPage):
            generic_select3: TestChoices

        user_input_1 = yield TestForm1

        if user_input_1.generic_select1 == TestChoices.A:
            user_input_2 = yield TestForm2
        else:
            user_input_2 = yield TestForm3

        return {**user_input_1.dict(), **user_input_2.dict()}

    # Submit 1
    form = generate_form(input_form, {"previous": True}, [])

    assert form == {
        "title": "Some title",
        "type": "object",
        "additionalProperties": False,
        "definitions": {
            "TestChoices": {
                "description": "An enumeration.",
                "enum": ["a", "b"],
                "title": "TestChoices",
                "type": "string",
            }
        },
        "properties": {"generic_select1": {"$ref": "#/definitions/TestChoices"}},
        "required": ["generic_select1"],
    }

    # Submit 2
    form = generate_form(input_form, {"previous": True}, [{"generic_select1": "b"}])

    assert form == {
        "title": "unknown",
        "type": "object",
        "additionalProperties": False,
        "definitions": {
            "TestChoices": {
                "description": "An enumeration.",
                "enum": ["a", "b"],
                "title": "TestChoices",
                "type": "string",
            }
        },
        "properties": {"generic_select3": {"$ref": "#/definitions/TestChoices"}},
        "required": ["generic_select3"],
    }

    # Submit complete
    form = generate_form(input_form, {"previous": True}, [{"generic_select1": "b"}, {"generic_select3": "a"}])
    assert form is None



import asyncio
from concurrent.futures import Future
from http import HTTPStatus
from threading import Event
from time import sleep
from unittest import mock
from uuid import uuid4

import pytest

from orchestrator.config.assignee import Assignee
from orchestrator.db import EngineSettingsTable, ProcessStepTable, ProcessTable, db
from orchestrator.forms import FormValidationError
from orchestrator.services.processes import (
    SYSTEM_USER,
    _async_resume_processes,
    _db_create_process,
    _db_log_process_ex,
    _db_log_step,
    _run_process_async,
    _safe_logstep,
    start_process,
)
from orchestrator.settings import app_settings
from orchestrator.targets import Target
from orchestrator.utils.errors import ApiException, error_state_to_dict
from orchestrator.workflow import (
    Abort,
    Complete,
    Failed,
    ProcessStat,
    ProcessStatus,
    Skipped,
    StepStatus,
    Success,
    Suspend,
    Waiting,
    make_step_function,
    make_workflow,
    step,
)


def test_db_create_process():
    pid = uuid4()
    workflow = make_workflow(lambda: None, "wf description", None, Target.SYSTEM, [])
    workflow.name = "wf name"
    pstat = ProcessStat(pid, workflow, None, None, current_user="user")

    _db_create_process(pstat)

    process = ProcessTable.query.get(pid)
    assert process
    assert process.workflow == "wf name"
    assert process.is_task


def test_process_log_db_step_success():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Success(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "success"
    assert psteps[0].pid == pid
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.RUNNING
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 2
    assert psteps[1].status == "success"
    assert psteps[1].pid == pid
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.RUNNING
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_skipped():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Skipped(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "skipped"
    assert psteps[0].pid == pid
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.RUNNING
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 2
    assert psteps[1].status == "skipped"
    assert psteps[1].pid == pid
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.RUNNING
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_suspend():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Suspend(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "suspend"
    assert psteps[0].pid == pid
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.SUSPENDED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 2
    assert psteps[1].status == "suspend"
    assert psteps[1].pid == pid
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.SUSPENDED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_waiting():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = "expected failure"
    state = Waiting(error_state_to_dict(Exception(state_data)))

    result = _db_log_step(pstat, step, state)
    assert result.iswaiting()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "waiting"
    assert psteps[0].pid == pid
    assert psteps[0].state["error"] == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.WAITING
    assert p.last_step == "step"
    assert p.failed_reason == state_data
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert result.iswaiting()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "waiting"
    assert psteps[0].pid == pid
    pstep_state = psteps[0].state
    assert len(pstep_state["executed_at"]) == 1
    del pstep_state["executed_at"]
    assert pstep_state == {"class": "Exception", "error": state_data, "retries": 1}
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.WAITING
    assert p.last_step == "step"
    assert p.failed_reason == state_data
    assert p.assignee == "assignee"


def test_process_log_db_step_failed():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, Assignee.SYSTEM)

    state_data = Exception("Hard failure")
    state = Failed(state_data)

    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].pid == pid
    saved_state = error_state_to_dict(state_data)
    saved_state.pop("traceback")
    assert psteps[0].state == saved_state
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.FAILED
    assert p.last_step == "step"
    assert p.failed_reason == "Hard failure"
    assert p.assignee == Assignee.SYSTEM

    step = make_step_function(lambda: None, "step", None, Assignee.SYSTEM)
    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].pid == pid
    pstep_state = psteps[0].state
    assert len(pstep_state["executed_at"]) == 1
    del pstep_state["executed_at"]
    del pstep_state["class"]
    assert pstep_state == {"error": "Hard failure", "retries": 1}
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.FAILED
    assert p.last_step == "step"
    assert p.failed_reason == "Hard failure"
    assert p.assignee == Assignee.SYSTEM


def test_process_log_db_step_assertion_failed():
    pid = uuid4()
    p = ProcessTable(
        pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER, is_task=True
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")

    state_data = AssertionError("Assertion failure")
    state = Failed(state_data)

    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].pid == pid
    saved_state = error_state_to_dict(state_data)
    saved_state.pop("traceback")
    assert psteps[0].state == saved_state
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.INCONSISTENT_DATA
    assert p.last_step == "step"
    assert p.failed_reason == "Assertion failure"
    assert p.assignee == Assignee.NOC

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].pid == pid
    pstep_state = psteps[0].state
    assert len(pstep_state["executed_at"]) == 1
    del pstep_state["executed_at"]
    del pstep_state["class"]
    assert pstep_state == {"error": "Assertion failure", "retries": 1}
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.INCONSISTENT_DATA
    assert p.last_step == "step"
    assert p.failed_reason == "Assertion failure"
    assert p.assignee == Assignee.NOC


def test_process_log_db_step_api_failed():
    pid = uuid4()
    p = ProcessTable(
        pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER, is_task=True
    )
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")

    state_data = ApiException(HTTPStatus.BAD_GATEWAY, "API failure")
    state = Failed(state_data)

    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].pid == pid
    saved_state = error_state_to_dict(state_data)
    saved_state.pop("traceback")
    assert psteps[0].state == saved_state
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.API_UNAVAILABLE
    assert p.last_step == "step"
    assert p.failed_reason == "API failure"
    assert p.assignee == Assignee.SYSTEM

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state.on_failed(error_state_to_dict))
    assert result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "failed"
    assert psteps[0].pid == pid
    pstep_state = psteps[0].state
    assert len(pstep_state["executed_at"]) == 1
    del pstep_state["executed_at"]
    del pstep_state["class"]
    assert pstep_state == {
        "error": "API failure",
        "retries": 1,
        "headers": "",
        "status_code": HTTPStatus.BAD_GATEWAY,
        "body": None,
    }
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.API_UNAVAILABLE
    assert p.last_step == "step"
    assert p.failed_reason == "API failure"
    assert p.assignee == Assignee.SYSTEM


def test_process_log_db_step_abort():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Abort(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "abort"
    assert psteps[0].pid == pid
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.ABORTED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 2
    assert psteps[1].status == "abort"
    assert psteps[1].pid == pid
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.ABORTED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_complete():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Complete(state_data)

    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 1
    assert psteps[0].status == "complete"
    assert psteps[0].pid == pid
    assert psteps[0].state == state_data
    assert psteps[0].created_by == "user"
    assert p.last_status == ProcessStatus.COMPLETED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"

    step = make_step_function(lambda: None, "step", None, "assignee")
    result = _db_log_step(pstat, step, state)
    assert not result.isfailed()

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).all()

    assert len(psteps) == 2
    assert psteps[1].status == "complete"
    assert psteps[1].pid == pid
    assert psteps[1].state == state_data
    assert psteps[1].created_by == "user"
    assert p.last_status == ProcessStatus.COMPLETED
    assert p.last_step == "step"
    assert p.failed_reason is None
    assert p.assignee == "assignee"


def test_process_log_db_step_no_pid():
    pid = uuid4()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"foo": "bar"}
    state = Complete(state_data)

    with pytest.raises(ValueError) as exc_info:
        _db_log_step(pstat, step, state)
    assert f"Failed to write failure step to process: process with PID {pid} not found" in str(exc_info.value)


def test_process_log_db_step_deduplication():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")

    step1 = make_step_function(lambda: None, "step1", None, Assignee.SYSTEM)
    step2 = make_step_function(lambda: None, "step2", None, Assignee.SYSTEM)
    error_state_data = {"error": "Soft failure"}
    _db_log_step(pstat, step1, Failed(error_state_data))
    _db_log_step(pstat, step2, Failed(error_state_data))
    _db_log_step(pstat, step1, Failed(error_state_data))
    _db_log_step(pstat, step1, Success({}))
    _db_log_step(pstat, step1, Failed(error_state_data))

    psteps = ProcessStepTable.query.filter_by(pid=p.pid).order_by(ProcessStepTable.executed_at.asc()).all()

    assert psteps[0].name == "step1"
    assert psteps[0].status == "failed"
    assert psteps[0].state == error_state_data
    assert psteps[0].created_by == "user"

    assert psteps[1].name == "step2"
    assert psteps[1].status == "failed"
    assert psteps[1].state == error_state_data
    assert psteps[1].created_by == "user"

    assert psteps[2].name == "step1"
    assert psteps[2].status == "failed"
    assert psteps[2].state == error_state_data
    assert psteps[2].created_by == "user"

    assert psteps[3].name == "step1"
    assert psteps[3].status == "success"
    assert psteps[3].state == {}
    assert psteps[3].created_by == "user"

    assert psteps[4].name == "step1"
    assert psteps[4].status == "failed"
    assert psteps[4].state == error_state_data
    assert psteps[4].created_by == "user"

    assert len(psteps) == 5

    assert p.last_status == ProcessStatus.FAILED
    assert p.last_step == "step1"
    assert p.failed_reason == "Soft failure"
    assert p.assignee == Assignee.SYSTEM


def test_safe_logstep():
    pid = uuid4()
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"not_serializable": object()}
    state = Complete(state_data)

    with mock.patch("orchestrator.services.processes._db_log_step", spec=_db_log_step) as mock__db_log_step:

        mock__db_log_step.side_effect = [
            Exception("Failed to commit because of json serializable failure"),
            mock.sentinel.result,
        ]
        result = _safe_logstep(pstat, step, state)

        assert result == mock.sentinel.result
        mock__db_log_step.assert_has_calls(
            [
                mock.call(
                    pstat,
                    step,
                    state,
                    None,  # broadcast_func, which is None for in-memory broadcasting
                ),
                mock.call(
                    pstat,
                    step,
                    Failed(
                        {
                            "class": "Exception",
                            "error": "Failed to commit because of json serializable failure",
                            "traceback": mock.ANY,
                        }
                    ),
                    None,  # broadcast_func, which is None for in-memory broadcasting
                ),
            ]
        )


def test_safe_logstep_critical_failure():
    pid = uuid4()

    # Skip storing the actual process to let `_db_log_step` fail

    pstat = ProcessStat(pid, None, None, None, current_user="user")
    step = make_step_function(lambda: None, "step", None, "assignee")
    state_data = {"not_serializable": object()}
    state = Complete(state_data)

    with mock.patch(  # Mock just to be able to spy on calls
        "orchestrator.services.processes._db_log_step", spec=_db_log_step, side_effect=_db_log_step
    ) as mock__db_log_step:
        with pytest.raises(ValueError) as e:
            result = _safe_logstep(pstat, step, state)

            assert result == mock.sentinel.result

            mock__db_log_step.assert_has_calls(
                [
                    mock.call(pstat, step, state),
                    mock.call(
                        pstat,
                        step,
                        Failed(
                            {
                                "class": "Exception",
                                "error": f"Failed to write failure step to process: process with PID {pid} not found",
                                "traceback": mock.ANY,
                            }
                        ),
                    ),
                ]
            )

        assert f"Failed to write failure step to process: process with PID {pid} not found" in str(e.value)


@mock.patch("orchestrator.services.processes._get_process")
@mock.patch("orchestrator.services.processes.resume_process")
def test_async_resume_processes(mock_resume_process, mock_get_process, caplog):
    """Test that _async_resume_process() rejects running processes and handles failures."""
    processes = [mock.Mock(spec=ProcessTable()), mock.Mock(spec=ProcessTable()), mock.Mock(spec=ProcessTable())]
    processes[0].pid, processes[0].last_status = 123, ProcessStatus.RUNNING
    processes[1].pid, processes[1].last_status = 124, ProcessStatus.FAILED
    processes[2].pid, processes[2].last_status = 125, ProcessStatus.API_UNAVAILABLE

    # get_process() should be called 3 times
    mock_get_process.side_effect = processes

    # resume_process() should be called 2 times for the non-running processes; let 1 call fail
    mock_resume_process.side_effect = [None, ValueError("This workflow cannot be resumed")]

    # Don't set app_settings.TESTING=False because we want to await the result
    asyncio.run(_async_resume_processes(processes, "testusername"))

    assert len(mock_get_process.mock_calls) == 3
    assert len(mock_resume_process.mock_calls) == 2
    assert "Cannot resume a running process" in caplog.text  # pid 123 should not be resumed
    assert "Failed to resume process" in caplog.text  # pid 125 should fail
    assert "Completed resuming processes" in caplog.text


def test_db_log_process_ex():
    pid = uuid4()

    # No exceptions!
    _db_log_process_ex(pid, ValueError())

    # Now with existing process
    p = ProcessTable(pid=pid, workflow="workflow_key", last_status=ProcessStatus.CREATED, created_by=SYSTEM_USER)
    db.session.add(p)
    db.session.commit()

    _db_log_process_ex(pid, ValueError())

    assert p.last_status == "failed"


def test_run_process_async_success():
    pid = uuid4()
    event = Event()

    def run_func():
        return Success(event.wait(1))

    # Disable Testing setting since we want to run async
    app_settings.TESTING = False
    _run_process_async(pid, run_func)
    sleep(0.01)

    assert EngineSettingsTable.query.one().running_processes == 1

    event.set()
    sleep(1)

    assert EngineSettingsTable.query.one().running_processes == 0
    app_settings.TESTING = True


@mock.patch("orchestrator.services.processes._db_log_process_ex")
def test_run_process_async_exception(mock_db_log_process_ex):
    pid = uuid4()
    event = Event()

    def run_func():
        event.wait(1)
        raise ValueError("Failed")

    # Disable Testing setting since we want to run async
    app_settings.TESTING = False
    _run_process_async(pid, run_func)
    sleep(0.1)

    assert EngineSettingsTable.query.one().running_processes == 1

    event.set()
    sleep(0.1)

    assert EngineSettingsTable.query.one().running_processes == 0

    mock_db_log_process_ex.assert_called_once_with(pid, mock.ANY)
    assert repr(mock_db_log_process_ex.call_args[0][1]) == "ValueError('Failed')"
    app_settings.TESTING = True


@mock.patch("orchestrator.services.processes._run_process_async", return_value=(mock.sentinel.pid, Future()))
@mock.patch("orchestrator.services.processes._db_create_process")
@mock.patch("orchestrator.services.processes.post_process")
@mock.patch("orchestrator.services.processes.get_workflow")
def test_start_process(mock_get_workflow, mock_post_process, mock_db_create_process, mock_run_process_async):
    @step("test step")
    def test_step():
        pass

    wf = make_workflow(lambda: None, "description", None, Target.SYSTEM, [test_step])
    wf.name = "name"
    mock_get_workflow.return_value = wf

    mock_post_process.return_value = {"a": 1}

    result, _ = start_process(mock.sentinel.wf_name, [{"a": 2}], mock.sentinel.user)

    pstat = mock_db_create_process.call_args[0][0]
    assert result == mock.sentinel.pid
    assert pstat.current_user == mock.sentinel.user
    assert pstat.state.status == StepStatus.SUCCESS
    assert pstat.workflow == wf
    assert pstat.state.unwrap() == {
        "process_id": mock.ANY,
        "reporter": mock.sentinel.user,
        "workflow_name": mock.sentinel.wf_name,
        "workflow_target": Target.SYSTEM,
        "a": 1,
    }
    assert pstat.log == [test_step]
    mock_post_process.assert_called_once_with(
        mock.ANY,
        {
            "process_id": mock.ANY,
            "reporter": mock.sentinel.user,
            "workflow_name": mock.sentinel.wf_name,
            "workflow_target": Target.SYSTEM,
        },
        [{"a": 2}],
    )
    mock_get_workflow.assert_called_once_with(mock.sentinel.wf_name)

    mock_post_process.reset_mock()
    mock_post_process.side_effect = FormValidationError("", [])

    with pytest.raises(FormValidationError):
        start_process(mock.sentinel.wf_name, None, mock.sentinel.user)
    mock_post_process.assert_called_once_with(
        mock.ANY,
        {
            "process_id": mock.ANY,
            "reporter": mock.sentinel.user,
            "workflow_name": mock.sentinel.wf_name,
            "workflow_target": Target.SYSTEM,
        },
        [{}],
    )



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from inspect import isgeneratorfunction
from typing import Callable, Dict, Optional, cast
from uuid import UUID

from more_itertools import first_true
from pydantic import validator

from orchestrator.db import ProductTable, SubscriptionTable
from orchestrator.forms import FormPage
from orchestrator.forms.validators import ProductId
from orchestrator.services import subscriptions
from orchestrator.settings import app_settings
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, InputForm, InputStepFunc, State, StateInputStepFunc
from orchestrator.utils.state import form_inject_args
from orchestrator.workflow import StepList, Workflow, conditional, done, init, make_workflow, step
from orchestrator.workflows.steps import (
    cache_domain_models,
    remove_domain_model_from_cache,
    resync,
    store_process_subscription,
    unsync_unchecked,
)


def _generate_new_subscription_form(workflow_target: str, workflow_name: str) -> InputForm:
    class NewProductPage(FormPage):
        product: ProductId

        @validator("product", allow_reuse=True)
        def product_validator(cls, v: UUID) -> UUID:
            """Run validator for initial_input_forms to check if the product exists and that this workflow is valid to run for this product."""
            product = ProductTable.query.get(v)
            if product is None:
                raise ValueError("Product not found")

            # Check if there is no reason to prevent this workflow
            current_workflow = product.create_subscription_workflow_key()

            if current_workflow != workflow_name:
                raise ValueError("This workflow is not valid for this product")

            return v

    return NewProductPage


def wrap_create_initial_input_form(initial_input_form: Optional[InputStepFunc]) -> Optional[StateInputStepFunc]:
    """Wrap initial input for for create workflows.

    This is needed because the frontend expects all create workflows to start with a page that only contains the product.
    It also expects the second page to have some user visible inputs and the product *again*.
    """

    def create_initial_input_form_generator(state: State) -> FormGenerator:
        workflow_target: str = state["workflow_target"]
        workflow_name: str = state["workflow_name"]

        product_user_input = yield _generate_new_subscription_form(workflow_target, workflow_name)

        product = ProductTable.query.get(product_user_input.product)

        begin_state = {"product": product.product_id, "product_name": product.name}

        if initial_input_form is None:
            return begin_state

        form = form_inject_args(initial_input_form)({**state, **begin_state})

        if isgeneratorfunction(initial_input_form):
            user_input = yield from cast(FormGenerator, form)
        else:
            user_input_model = yield cast(InputForm, form)
            user_input = user_input_model.dict()

        return {**begin_state, **user_input}

    return create_initial_input_form_generator


# TODO these translations should stay in the frontend See also #549
TRANSLATIONS = {
    "subscription.not_in_sync": "Subscription already has a running process or task",
    "subscription.relations_not_in_sync": "This subscription can not be modified because some related subscriptions are not insync",
    "subscription.no_modify_invalid_status": "This subscription can not be modified because of the status it has",
    # no_modify_parent_subscription deprecated since "0.4.0", renamed to no_modify_subscription_in_use_by_others
    "subscription.no_modify_parent_subscription": "This subscription can not be modified as it is used in other subscriptions",
    "subscription.no_modify_subscription_in_use_by_others": "This subscription can not be modified as it is used in other subscriptions",
    "subscription.no_modify_auto_negotiation": "This workflow is not valid for this subscription",
}


def _generate_modify_form(workflow_target: str, workflow_name: str) -> InputForm:
    class ModifySubscriptionPage(FormPage):
        # We use UUID instead of SubscriptionId here because we dont want the allowed_status check and
        # we do our own validation here..
        subscription_id: UUID

        @validator("subscription_id", allow_reuse=True)
        def subscription_validator(cls, v: UUID, values: Dict) -> UUID:
            """Run validator for initial_input_forms to check if the subscription exists and that this workflow is valid to run for this subscription."""
            subscription = SubscriptionTable.query.get(v)
            if subscription is None:
                raise ValueError("Subscription not found")

            # Check if there is no reason to prevent this workflow
            workflows = subscriptions.subscription_workflows(subscription)
            current_workflow = first_true(
                workflows[workflow_target.lower()], None, lambda wf: wf["name"] == workflow_name
            )

            if not current_workflow:
                raise ValueError("This workflow is not valid for this subscription")

            if "reason" in current_workflow:
                message = TRANSLATIONS.get(current_workflow["reason"], current_workflow["reason"])
                raise ValueError(f"This workflow cannot be started: {message}")

            return v

    return ModifySubscriptionPage


def wrap_modify_initial_input_form(initial_input_form: Optional[InputStepFunc]) -> Optional[StateInputStepFunc]:
    """Wrap initial input for modify and terminate workflows.

    This is needed because the frontend expects all modify workflows to start with a page that only contains the
    subscription id. It also expects the second page to have some user visible inputs and the subscription id *again*.
    """

    def create_initial_input_form_generator(state: State) -> FormGenerator:
        workflow_target: str = state["workflow_target"]
        workflow_name: str = state["workflow_name"]

        user_input = yield _generate_modify_form(workflow_target, workflow_name)

        subscription = SubscriptionTable.query.get(user_input.subscription_id)

        begin_state = {
            "subscription_id": str(subscription.subscription_id),
            "product": str(subscription.product_id),
            "organisation": str(subscription.customer_id),
        }

        if initial_input_form is None:
            return begin_state

        form = form_inject_args(initial_input_form)({**state, **begin_state})

        if isgeneratorfunction(initial_input_form):
            user_input = yield from cast(FormGenerator, form)
        else:
            user_input_model = yield cast(InputForm, form)
            user_input = user_input_model.dict()

        return {**begin_state, **user_input}

    return create_initial_input_form_generator


modify_initial_input_form_generator = None


validate_initial_input_form_generator = wrap_modify_initial_input_form(modify_initial_input_form_generator)


def validate_workflow(description: str) -> Callable[[Callable[[], StepList]], Workflow]:
    """Transform an initial_input_form and a step list into a workflow.

    Use this for subscription validate workflows.

    Example::

        @validate_workflow("create service port")
        def create_service_port():
            do_something
            >> do_something_else
    """

    def _validate_workflow(f: Callable[[], StepList]) -> Workflow:
        push_subscriptions = conditional(lambda _: app_settings.CACHE_DOMAIN_MODELS)
        steplist = (
            init
            >> store_process_subscription(Target.SYSTEM)
            >> push_subscriptions(remove_domain_model_from_cache)
            >> unsync_unchecked
            >> f()
            >> resync
            >> push_subscriptions(cache_domain_models)
            >> done
        )

        return make_workflow(f, description, validate_initial_input_form_generator, Target.SYSTEM, steplist)

    return _validate_workflow


@step("Equalize workflow step count")
def obsolete_step() -> None:
    """Equalize workflow step counts.

    When changing existing workflows that might have suspended instances in production it is important
    to keep the exact same amount of steps. This step can be used to fill the gap when removing a step.
    """
    pass



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import operator
from collections import abc
from functools import reduce
from typing import AbstractSet, Any, Iterable, Iterator, List, Optional, Sequence, Tuple, Union, cast

from more_itertools import first, last

from orchestrator.utils.functional import expand_ranges, to_ranges


class VlanRanges(abc.Set):
    """Represent VLAN ranges.

    This class is quite liberal in what it accepts as valid VLAN ranges. All of:

    - overlapping ranges
    - ranges with start value > stop value
    - ranges with extraneous whitespace

    are all accepted and normalized to a canonical value.

    Examples::

        # These are all equivalent
        VlanRanges("4,10-12,11-14")
        VlanRanges("4,  ,11 - 14, 10-  12")
        VlanRanges("4,10-14")
        VlanRanges([4, 10, 11, 12, 13, 14])
        VlanRanges([[4], [10,12], [11,14]])
        VlanRanges([(4, 4), (10, 14)])

    """

    _vlan_ranges: Tuple[range, ...]

    def __init__(self, val: Optional[Union[str, int, Iterable[int], Sequence[Sequence[int]]]] = None) -> None:
        # The idea is to bring all acceptable values to one canonical intermediate format: the `Sequence[Sequence[
        # int]]`. Where the inner sequence is either a one or two element sequence. The one element sequence
        # represents a single VLAN, the two element sequence represents a VLAN range.
        #
        # An example of this intermediate format is::
        #
        #     vlans = [[5], [10, 12]]
        #
        # That example represents 4 VLANs, namely: 5, 10, 11, 12. The latter three VLANs are encode as a range.
        #
        # This intermediate format happens to be the format as accepted by :func:`expand_ranges`. This function has
        # the advantage of deduplicating overlapping ranges or VLANs specified more than once. In addition its return
        # value can be use as input to the :func:`to_ranges` function.
        vlans: Sequence[Sequence[int]] = []
        if val is None:
            self._vlan_ranges = ()
            return
        elif isinstance(val, str):
            if val.strip() != "":
                # This might look complex, but it does handle strings such as `"  3, 4, 6-9, 4, 8 - 10"`
                try:
                    vlans = list(map(lambda s: list(map(int, s.strip().split("-"))), val.split(",")))
                except ValueError:
                    raise ValueError(f"{val} could not be converted to a {self.__class__.__name__} object.")
        elif isinstance(val, int):
            vlans = [[val]]
        elif isinstance(val, abc.Sequence):
            if len(val) > 0:
                if isinstance(first(val), int):
                    vlans = list(map(lambda x: [x], val))
                elif isinstance(first(val), abc.Sequence):
                    vlans = cast(Sequence[Sequence[int]], val)
                else:
                    raise ValueError(f"{val} could not be converted to a {self.__class__.__name__} object.")
        elif isinstance(val, abc.Iterable):
            vlans = list(map(lambda x: [x], val))  # type: ignore
        else:
            raise ValueError(f"{val} could not be converted to a {self.__class__.__name__} object.")

        er = expand_ranges(vlans, inclusive=True)
        if er and not (first(er) >= 0 and last(er) <= 4096):
            raise ValueError(f"{val} is out of range (0-4096).")

        self._vlan_ranges = tuple(to_ranges(er))

    def to_list_of_tuples(self) -> List[Tuple[int, int]]:
        """Construct list of tuples representing the VLAN ranges.

        Example::

            >>> VlanRanges("10 - 12, 8").to_list_of_tuples()
            [(8, 8), (10, 12)]

        Returns:
            The VLAN ranges as contained in this object.

        """
        # `range` objects have an exclusive `stop`. VlanRanges is expressed using terms that use an inclusive stop,
        # which is one less then the exclusive one we use for the internal representation. Hence the `-1`
        return [(vr.start, vr.stop - 1) for vr in self._vlan_ranges]

    def __contains__(self, key: object) -> bool:
        return any(map(lambda range_from_self: key in range_from_self, self._vlan_ranges))

    def __iter__(self) -> Iterator[int]:
        # The power of choosing proper abstractions: `range` objects already define an __iter__ method. Hence all we
        # need to do, is delegated to them.
        for vr in self._vlan_ranges:
            yield from vr

    def __len__(self) -> int:
        """Return the number of VLANs represented by this VlanRanges object.

        Returns:
            Number of VLAN's

        """
        return sum(len(r) for r in self._vlan_ranges)

    def __str__(self) -> str:
        # `range` objects have an exclusive `stop`. VlanRanges is expressed using terms that use an inclusive stop,
        # which is one less then the exclusive one we use for the internal representation. Hence the `-1`
        return ",".join(str(vr.start) if len(vr) == 1 else f"{vr.start}-{vr.stop - 1}" for vr in self._vlan_ranges)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self.to_list_of_tuples())})"

    def __json__(self) -> str:
        return str(self)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        return self._vlan_ranges == o._vlan_ranges

    def __hash__(self) -> int:
        return hash(self._vlan_ranges)

    def __sub__(self, other: Union[int, AbstractSet[Any]]) -> VlanRanges:
        if isinstance(other, int):
            new_set = set(self)
            new_set.remove(other)
            return VlanRanges(new_set)
        else:
            return VlanRanges(set(self) - set(other))

    def __and__(self, other: AbstractSet[Any]) -> VlanRanges:
        return VlanRanges(set(self) & set(other))

    def __or__(self, other: AbstractSet[Any]) -> VlanRanges:
        return VlanRanges(set(self) | set(other))

    def __xor__(self, other: AbstractSet[Any]) -> VlanRanges:
        return VlanRanges(set(self) ^ set(other))

    def isdisjoint(self, other: Iterable[Any]) -> bool:
        return set(self).isdisjoint(other)

    def union(self, *others: AbstractSet[Any]) -> VlanRanges:
        return reduce(operator.__or__, others, self)

    @classmethod
    def __get_validators__(cls) -> Iterator:
        yield cls.validate

    @classmethod
    def validate(cls, v: Optional[Union[str, int, Iterable[int], Sequence[Sequence[int]]]]) -> VlanRanges:
        # The constructor of VlanRanges performs the validations.
        if isinstance(v, VlanRanges):
            return v
        return cls(v)



import time
from http import HTTPStatus
from threading import Condition
from typing import Any, Dict, Generator
from uuid import uuid4

import pytest
from fastapi import status
from fastapi.websockets import WebSocketDisconnect

from orchestrator.config.assignee import Assignee
from orchestrator.db import ProcessStepTable, ProcessSubscriptionTable, ProcessTable, WorkflowTable, db
from orchestrator.settings import app_settings
from orchestrator.targets import Target
from orchestrator.types import UUIDstr
from orchestrator.utils.json import json_loads
from orchestrator.websocket import websocket_manager
from orchestrator.workflow import ProcessStatus, StepStatus, done, init, step, workflow
from test.unit_tests.workflows import WorkflowInstanceForTests

test_condition = Condition()

LONG_RUNNING_STEP = "Long Running Step"
IMMEDIATE_STEP = "Immediate Step"


@pytest.fixture
def long_running_workflow():
    @step(LONG_RUNNING_STEP)
    def long_running_step():
        with test_condition:
            test_condition.wait()
        return {"done": True}

    @step(IMMEDIATE_STEP)
    def immediate_step():
        return {"done": True}

    @workflow("Long Running Workflow")
    def long_running_workflow_py():
        return init >> long_running_step >> immediate_step >> long_running_step >> done

    with WorkflowInstanceForTests(long_running_workflow_py, "long_running_workflow_py"):

        db_workflow = WorkflowTable(name="long_running_workflow_py", target=Target.MODIFY)
        db.session.add(db_workflow)
        db.session.commit()

        yield "long_running_workflow_py"


@pytest.fixture
def test_workflow_2(generic_subscription_1: UUIDstr, generic_product_type_1) -> Generator:
    _, GenericProductOne = generic_product_type_1

    @step("Insert UUID in state")
    def insert_object():
        return {"subscription_id": str(uuid4()), "model": GenericProductOne.from_subscription(generic_subscription_1)}

    @step("Test that it is a string now")
    def check_object(subscription_id: Any, model: Dict) -> None:
        # This is actually a test. It would be nicer to have this in a proper test but that takes to much setup that
        # already happens here. So we hijack this fixture and run this test for all tests that use this fixture
        # (which should not be an issue)
        assert isinstance(subscription_id, str)
        assert isinstance(model, dict)

    @step("immediate step")
    def immediate_step():
        return {"done": True}

    @workflow("Workflow")
    def workflow_for_testing_processes_2_py():
        return init >> insert_object >> check_object >> immediate_step >> done

    with WorkflowInstanceForTests(workflow_for_testing_processes_2_py, "workflow_for_testing_processes_2_py"):
        db_workflow = WorkflowTable(name="workflow_for_testing_processes_2_py", target=Target.MODIFY)
        db.session.add(db_workflow)
        db.session.commit()

        yield "workflow_for_testing_processes_2_py"


@pytest.fixture
def completed_process(test_workflow, generic_subscription_1):
    pid = uuid4()
    process = ProcessTable(pid=pid, workflow=test_workflow, last_status=ProcessStatus.COMPLETED)
    init_step = ProcessStepTable(pid=pid, name="Start", status=StepStatus.SUCCESS, state={})
    insert_step = ProcessStepTable(
        pid=pid,
        name="Insert UUID in state",
        status=StepStatus.SUCCESS,
        state={"subscription_id": generic_subscription_1},
    )
    check_step = ProcessStepTable(
        pid=pid,
        name="Test that it is a string now",
        status=StepStatus.SUCCESS,
        state={"subscription_id": generic_subscription_1},
    )
    step = ProcessStepTable(
        pid=pid, name="Modify", status=StepStatus.SUCCESS, state={"subscription_id": generic_subscription_1}
    )

    process_subscription = ProcessSubscriptionTable(pid=pid, subscription_id=generic_subscription_1)

    db.session.add(process)
    db.session.add(init_step)
    db.session.add(insert_step)
    db.session.add(check_step)
    db.session.add(step)
    db.session.add(process_subscription)
    db.session.commit()

    return pid


# long running workflow test only works locally and with memory type
def test_websocket_process_detail_workflow(test_client, long_running_workflow):
    if websocket_manager.broadcaster_type != "memory" or app_settings.ENVIRONMENT != "local":
        pytest.skip("test does not work with redis")

    app_settings.TESTING = False

    # Start the workflow
    response = test_client.post(f"/api/processes/{long_running_workflow}", json=[{}])
    assert (
        HTTPStatus.CREATED == response.status_code
    ), f"Invalid response status code (response data: {response.json()})"

    pid = response.json()["id"]

    response = test_client.get(f"api/processes/{pid}")
    assert HTTPStatus.OK == response.status_code

    # Make sure it started again
    time.sleep(1)

    try:
        with test_client.websocket_connect("api/processes/all?token=") as websocket:
            # Check the websocket messages.
            # the initial process details.
            data = websocket.receive_text()
            process = json_loads(data)["process"]
            assert process["workflow_name"] == "long_running_workflow_py"
            assert process["status"] == ProcessStatus.RUNNING

            # Let first long step finish, receive_text would otherwise wait for a message indefinitely.
            with test_condition:
                test_condition.notify_all()
            time.sleep(1)

            # message step 1.
            data = websocket.receive_text()
            json_data = json_loads(data)
            process = json_data["process"]
            assert process["status"] == ProcessStatus.RUNNING
            assert process["steps"][1]["name"] == LONG_RUNNING_STEP
            assert process["steps"][1]["status"] == StepStatus.SUCCESS

            # message step 2.
            data = websocket.receive_text()
            json_data = json_loads(data)
            process = json_data["process"]
            assert process["status"] == ProcessStatus.RUNNING
            assert process["steps"][2]["name"] == IMMEDIATE_STEP
            assert process["steps"][2]["status"] == StepStatus.SUCCESS

            # Let second long step finish, receive_text would otherwise wait for a message indefinitely.
            with test_condition:
                test_condition.notify_all()
            time.sleep(1)

            # message step 3.
            data = websocket.receive_text()
            json_data = json_loads(data)
            process = json_data["process"]
            assert process["status"] == ProcessStatus.RUNNING
            assert process["steps"][3]["name"] == LONG_RUNNING_STEP
            assert process["steps"][3]["status"] == StepStatus.SUCCESS

            # message step 4.
            data = websocket.receive_text()
            json_data = json_loads(data)
            process = json_data["process"]
            assert process["status"] == ProcessStatus.COMPLETED
            assert process["steps"][4]["name"] == "Done"
            assert process["steps"][4]["status"] == StepStatus.COMPLETE
            assert json_data["close"] is True

            # close and call receive_text to check websocket close exception
            websocket.close()
            data = websocket.receive_text()
    except WebSocketDisconnect as exception:
        assert exception.code == status.WS_1000_NORMAL_CLOSURE
    except AssertionError as e:
        # Finish steps so the test doesn't get stuck waiting forever.
        with test_condition:
            test_condition.notify_all()
        with test_condition:
            test_condition.notify_all()
        app_settings.TESTING = True
        raise e

    with test_condition:
        test_condition.notify_all()
    app_settings.TESTING = True


def test_websocket_process_detail_with_suspend(test_client, test_workflow):
    if websocket_manager.broadcaster_type != "memory":
        pytest.skip("test does not work with redis")

    response = test_client.post(f"/api/processes/{test_workflow}", json=[{}])

    assert (
        HTTPStatus.CREATED == response.status_code
    ), f"Invalid response status code (response data: {response.json()})"

    pid = response.json()["id"]

    try:
        with test_client.websocket_connect("api/processes/all?token=") as websocket:
            # Resume process
            user_input = {"generic_select": "A"}

            response = test_client.put(f"/api/processes/{pid}/resume", json=[user_input])
            assert HTTPStatus.NO_CONTENT == response.status_code

            data = websocket.receive_text()
            process = json_loads(data)["process"]
            assert process["status"] == ProcessStatus.SUSPENDED
            assert process["assignee"] == Assignee.CHANGES

            data = websocket.receive_text()
            process = json_loads(data)["process"]
            assert process["status"] == ProcessStatus.RUNNING
            assert process["assignee"] == Assignee.CHANGES

            data = websocket.receive_text()
            process = json_loads(data)["process"]
            assert process["status"] == ProcessStatus.COMPLETED
            assert process["assignee"] == Assignee.SYSTEM

            # close and call receive_text to check websocket close exception
            websocket.close()
            data = websocket.receive_text()
    except WebSocketDisconnect as exception:
        assert exception.code == status.WS_1000_NORMAL_CLOSURE


def test_websocket_process_detail_with_abort(test_client, test_workflow):
    if websocket_manager.broadcaster_type != "memory":
        pytest.skip("test does not work with redis")

    response = test_client.post(f"/api/processes/{test_workflow}", json=[{}])

    assert (
        HTTPStatus.CREATED == response.status_code
    ), f"Invalid response status code (response data: {response.json()})"

    pid = response.json()["id"]

    try:
        with test_client.websocket_connect("api/processes/all?token=") as websocket:
            # Abort process
            response = test_client.put(f"/api/processes/{pid}/abort")
            assert HTTPStatus.NO_CONTENT == response.status_code

            data = websocket.receive_text()
            process = json_loads(data)["process"]
            assert process["status"] == ProcessStatus.SUSPENDED
            assert process["assignee"] == Assignee.CHANGES

            data = websocket.receive_text()
            process = json_loads(data)["process"]
            assert process["status"] == ProcessStatus.ABORTED
            assert process["assignee"] == Assignee.SYSTEM

            # close and call receive_text to check websocket close exception
            websocket.close()
            data = websocket.receive_text()
    except WebSocketDisconnect as exception:
        assert exception.code == status.WS_1000_NORMAL_CLOSURE


def test_websocket_process_list_multiple_workflows(test_client, test_workflow, test_workflow_2):
    # This tests the ProcessDataBroadcastThread as well
    if websocket_manager.broadcaster_type != "memory":
        pytest.skip("test does not work with redis")

    # to keep track of the amount of websocket messages
    message_count = 0

    expected_workflow_1_steps = [
        {
            "assignee": Assignee.SYSTEM,
            "status": ProcessStatus.RUNNING,
            "step": "Start",
            "step_status": StepStatus.SUCCESS,
        },
        {
            "assignee": Assignee.SYSTEM,
            "status": ProcessStatus.RUNNING,
            "step": "Insert UUID in state",
            "step_status": StepStatus.SUCCESS,
        },
        {
            "assignee": Assignee.SYSTEM,
            "status": ProcessStatus.RUNNING,
            "step": "Test that it is a string now",
            "step_status": StepStatus.SUCCESS,
        },
        {
            "assignee": Assignee.CHANGES,
            "status": ProcessStatus.SUSPENDED,
            "step": "Modify",
            "step_status": StepStatus.SUSPEND,
        },
    ]

    expected_workflow_2_steps = [
        {
            "assignee": Assignee.SYSTEM,
            "status": ProcessStatus.RUNNING,
            "step": "Start",
            "step_status": StepStatus.SUCCESS,
        },
        {
            "assignee": Assignee.SYSTEM,
            "status": ProcessStatus.RUNNING,
            "step": "Insert UUID in state",
            "step_status": StepStatus.SUCCESS,
        },
        {
            "assignee": Assignee.SYSTEM,
            "status": ProcessStatus.RUNNING,
            "step": "Test that it is a string now",
            "step_status": StepStatus.SUCCESS,
        },
        {
            "assignee": Assignee.SYSTEM,
            "status": ProcessStatus.RUNNING,
            "step": "immediate step",
            "step_status": StepStatus.SUCCESS,
        },
        {
            "assignee": Assignee.SYSTEM,
            "status": ProcessStatus.COMPLETED,
            "step": "Done",
            "step_status": StepStatus.COMPLETE,
        },
    ]

    test_workflow_1_messages = []
    test_workflow_2_messages = []

    try:
        with test_client.websocket_connect("api/processes/all/?token=") as websocket:
            # start test_workflow
            test_workflow_response = test_client.post(f"/api/processes/{test_workflow}", json=[{}])

            assert (
                HTTPStatus.CREATED == test_workflow_response.status_code
            ), f"Invalid response status code (response data: {test_workflow_response.json()})"

            test_workflow_1_pid = test_workflow_response.json()["id"]

            # Start test_workflow_2
            response = test_client.post(f"/api/processes/{test_workflow_2}", json=[{}])
            assert (
                HTTPStatus.CREATED == response.status_code
            ), f"Invalid response status code (response data: {response.json()})"

            test_workflow_2_pid = response.json()["id"]

            # Make sure it started again
            time.sleep(1)

            # close and call receive_text to check websocket close exception
            websocket.close()

            # Checks if the correct messages are send, without order for which workflow.
            while True:
                if message_count == 9:
                    break
                message = websocket.receive_text()
                message_count += 1
                json_data = json_loads(message)
                assert "process" in json_data, f"Websocket message does not contain process: {json_data}"
                process_id = json_data["process"]["id"]

                if process_id == test_workflow_1_pid:
                    test_workflow_1_messages.append(json_data)
                elif process_id == test_workflow_2_pid:
                    test_workflow_2_messages.append(json_data)
    except WebSocketDisconnect as exception:
        assert exception.code == status.WS_1000_NORMAL_CLOSURE
    except AssertionError as e:
        raise e

    assert message_count == 9

    for index, message in enumerate(test_workflow_1_messages):
        process = message["process"]
        expectedData = expected_workflow_1_steps.pop(0)

        assert process["assignee"] == expectedData["assignee"]
        assert process["status"] == expectedData["status"]

        assert process["step"] == expectedData["step"]
        assert process["steps"][index]["name"] == expectedData["step"]
        assert process["steps"][index]["status"] == expectedData["step_status"]

    for index, message in enumerate(test_workflow_2_messages):
        process = message["process"]
        expectedData = expected_workflow_2_steps.pop(0)

        assert process["assignee"] == expectedData["assignee"]
        assert process["status"] == expectedData["status"]

        assert process["step"] == expectedData["step"]
        assert process["steps"][index]["name"] == expectedData["step"]
        assert process["steps"][index]["status"] == expectedData["step_status"]



import pytest

from orchestrator.db import FixedInputTable, ProductTable, db
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import ProductModel, SubscriptionModel
from orchestrator.domain.lifecycle import ProductLifecycle
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_type_one(test_product_block_one):
    ProductBlockOneForTestInactive, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one

    class ProductTypeOneForTestInactive(SubscriptionModel, is_base=True):
        test_fixed_input: bool
        block: ProductBlockOneForTestInactive

    class ProductTypeOneForTestProvisioning(
        ProductTypeOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        test_fixed_input: bool
        block: ProductBlockOneForTestProvisioning

    class ProductTypeOneForTest(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTest
    yield ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest
    del SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"]


@pytest.fixture
def test_product_one(test_product_block_one_db):
    product = ProductTable(
        name="TestProductOne", description="Test ProductTable", product_type="Test", tag="TEST", status="active"
    )

    fixed_input = FixedInputTable(name="test_fixed_input", value="False")

    product_block, _ = test_product_block_one_db
    product.fixed_inputs = [fixed_input]
    product.product_blocks = [product_block]

    db.session.add(product)
    db.session.commit()

    return product.product_id


@pytest.fixture
def test_product_model(test_product_one):
    product = ProductTable.query.filter(ProductTable.product_id == test_product_one).one()

    return ProductModel(
        product_id=test_product_one,
        name="TestProductOne",
        description="Test ProductTable",
        product_type="Test",
        tag="TEST",
        status=ProductLifecycle.ACTIVE,
        created_at=product.created_at,
    )



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for custom serialization to and from JSON.

Both :func:`json.dumps` and :func:`json.loads` provide hooks for encoding and decoding objects they normally can't
encode and decode. Their hooks operate somewhat differently though.

:func:`json.dumps` has a `default` keyword argument that can be supplied with a function that will **only** be called
for objects it cannot encode. This function should return a new Python object (not a str in JSON format!) that can be
serialized to JSON by the active encoder. Generally this means that we should convert custom data types to either a
`str`, `int`, `float`, `bool` or a `dict` with only these primitive values. As the `default` function is a matter of
last resort it should raise a :exc:`TypeError` if it too can't encode an object. We have named the `default` function
:func:`to_serialize`.

To ensure we can decode what we have encoded we will include extra information about the custom data types, namely:

- the module the data type resides in
- the name of the data type

This information is encoded under the keys `"__module__"` and `"__class__"` respectively.

Upon decoding this information will be used to import the module and instantiate the class for the the custom data
type dynamically. One obviously has to encode whatever information that's necessary to be able to decode it later as
well.

.. note:: There are libraries that provide this functionally out of the box. We have chosen not to use them as their
   means of encoding dumps the _internal_ representation of custom types. This can lead to fairly unreadable JSON. Eg.
   encoding a :obj:`uuid.UUID` object results in a large integer value plus an additional attribute `is_safe` that has
   no meaning to us. We, instead, want something that's human readable as well.

.. warning:: Another thing to realize is that although we do encode :obj:`uuid.UUID` objects, we do not decode them.
   Past project decisions still haunt use and we generally process UUIDs as `str`s in our backend, only converting them
   to real :obj:`uuid.UUID` objects where necessary. Don't simply assume UUID decoding logic was forgotten; it was not!

.. note:: Even though we do serialize domain models for audit log purposes to the state, we do not deserialize from
   the state. Instead we look them up in the DB instead (see: func:`server.utils.state.inject_args`). The reason for
   this decision is that we don't want to break currently running workflows when data models have their definition
   changed. If the database is the serialization store, we can deal with changes by means of DB data migrations. This
   is much, much harder to do when the serialization store is a JSON blob.

:func:`json.loads`'s hook works somewhat differently. Whereas the `default` function of :func:`json.dumps` is only
called as a matter of last resort, the `object_hook` function of :func:`json.loads is called for every `dict` value!
This means that, instead of raising an exception, it should return the `dict` that was passed in if it can't decode it.

The `object_hook` function will be called with the inner most `dict` first, working its way up to the top level `dict`.
Eg. Given::

    json_str = '{"a": 1, "b": 2, "c": {"x": true, "y": false}}'

The `object_hook` function will be called twice. First with the `dict`::

    {'x': True, 'y': False}

And then with the `dict`::

    {'a': 1, 'b': 2, 'c': {'x': True, 'y': False}}

:func:`json.loads` will replace the `dict` that it passed to `object_hook` function, with the object that the
`object_hook` function returned.

We have named the `object_hook` function :func:`from_serializable`. Reason for naming these custom encoding and
decoding functions differently then `default` and `object_hook` is that they are also used outside of the context of
:func:`json.dumps` and :func:`json.loads`.

"""
import re
from contextlib import suppress
from dataclasses import asdict, is_dataclass
from datetime import datetime
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network
from typing import Any, Dict, List, Sequence, Set, Tuple, Union
from uuid import UUID

import orjson as json
import rapidjson as rjson
import structlog
from pydantic import BaseModel

from orchestrator.utils.datetime import isoformat

PY_JSON_TYPES = Union[Dict[str, Any], List, str, int, float, bool, None, object]  # pragma: no mutate

logger = structlog.get_logger(__name__)


def json_loads(s: Union[str, bytes, bytearray]) -> PY_JSON_TYPES:
    o = json.loads(s)
    if isinstance(o, list):
        return [from_serializable(dikt) for dikt in o]
    return from_serializable(o)


def json_dumps(obj: PY_JSON_TYPES) -> str:
    try:
        return json.dumps(
            obj,
            default=to_serializable,
            option=json.OPT_PASSTHROUGH_DATETIME | json.OPT_OMIT_MICROSECONDS | json.OPT_NON_STR_KEYS,
        ).decode("utf8")
    except TypeError as e:
        # When Recursion limit is not configurable in orjson, falling back to the next best lib.
        if str(e) == "default serializer exceeds recursion limit":
            return rjson.dumps(obj, default=to_serializable)
        else:
            raise e


def to_serializable(o: Any) -> Any:
    """Convert an object into an object that the JSON encode can serialize.

    Args:
        o: Object to convert.

    Returns:
        Serializable object.

    Raises:
        TypeError: in case no conversion was possible.

    """
    if isinstance(o, UUID):
        return str(o)
    if (
        isinstance(o, IPv4Address)
        or isinstance(o, IPv6Address)
        or isinstance(o, IPv4Network)
        or isinstance(o, IPv6Network)
    ):
        return str(o)
    if isinstance(o, datetime):
        return isoformat(o)
    if is_dataclass(o):
        return asdict(o)
    if hasattr(o, "__json__"):
        return o.__json__()
    if hasattr(o, "to_dict"):
        # api_client models all have a to_dict function
        return o.to_dict()
    if isinstance(o, BaseModel):
        return o.dict()
    if isinstance(o, Set):
        return list(o)
    raise TypeError(f"Could not serialize object of type {o.__class__.__name__} to JSON")


ISO_FORMAT_STR_LEN = len("2019-05-18T15:17:00+00:00")  # assume 'seconds' precision
UUID_PATTERN = re.compile(r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$", re.IGNORECASE)


def from_serializable(dct: Dict[str, Any]) -> Dict[str, Any]:
    """Convert a serializable object into a more specific, custom, Python data type.

    Args:
        dct: Serializable object.

    Returns:
        Either the serializable object itself or a more specific data type if a conversion was possible.

    """
    for k, v in dct.items():
        # We don't want to try converting each string we come across to a `datetime` object, hence we employ a simple
        # and reasonably specific heuristic that identifies likely ISO formatted string candidates
        if isinstance(v, str) and len(v) == ISO_FORMAT_STR_LEN and v[10] == "T":
            with suppress(ValueError, TypeError):
                timestamp = datetime.fromisoformat(v)
                assert timestamp.tzinfo is not None, "All timestamps should contain timezone information."  # noqa: S101
                dct[k] = timestamp
    return dct


def non_none_dict(dikt: Sequence[Tuple[str, Any]]) -> Dict[Any, Any]:
    """
    Return no `None` values in a Dict.

    This function may be used in the `asdict()` method as dictionary factory.

    Args:
        dikt: Tuple of values where

    Returns:
        dict of the keys and values if value is not None

    Examples:
        >>> non_none_dict({"a": 0, "b": None, "c": ""}.items())
        {'a': 0, 'c': ''}

    """
    return {k: v for k, v in dikt if v is not None}



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

from more_itertools import flatten
from more_itertools.more import one
from more_itertools.recipes import first_true
from pydantic import ValidationError
from sqlalchemy import not_
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.sqltypes import JSON

import orchestrator.workflows
from orchestrator.api.api_v1.endpoints.fixed_input import fi_configuration
from orchestrator.db import FixedInputTable, ProductTable, SubscriptionTable, WorkflowTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.services import products
from orchestrator.services.translations import generate_translations
from orchestrator.targets import Target
from orchestrator.types import State
from orchestrator.utils.errors import ProcessFailure
from orchestrator.workflow import StepList, done, init, step, workflow

# Since these errors are probably programming failures we should not throw AssertionErrors


@step("Check all workflows in database")
def check_all_workflows_are_in_db() -> State:
    all_workflows_in_db = {k.name for k in WorkflowTable.query.all()}
    all_workflows = {k for k in orchestrator.workflows.ALL_WORKFLOWS.keys()}  # noqa: C416
    not_in_db = all_workflows - all_workflows_in_db
    not_in_lwi = all_workflows_in_db - all_workflows
    if not_in_db or not_in_lwi:
        raise ProcessFailure(
            "Found missing workflows in database or implementations",
            {
                "Workflows not registered in the database": list(not_in_db),
                "Workflows not registered in a `LazyWorkflowInstance`": list(not_in_lwi),
            },
        )

    return {"check_all_workflows_are_in_db": True}


@step("Check workflows for matching targets and descriptions")
def check_workflows_for_matching_targets_and_descriptions() -> State:
    workflow_assertions = []
    for key, lazy_wf in orchestrator.workflows.ALL_WORKFLOWS.items():
        wf = lazy_wf.instantiate()
        db_workflow = WorkflowTable.query.filter(WorkflowTable.name == key).first()
        if db_workflow:
            # Test workflows might not exist in the database
            if (
                wf.target != db_workflow.target
                or wf.name != db_workflow.name
                or wf.description != db_workflow.description
            ):
                message = (
                    f"Workflow {wf.name}: {wf.target} <=> {db_workflow.target}, "
                    f"{wf.name} <=> {db_workflow.name} and {wf.description} <=> {db_workflow.description}. "
                )
                workflow_assertions.append(message)

    if workflow_assertions:
        workflow_message = "\n".join(workflow_assertions)
        raise ProcessFailure("Workflows with none matching targets and descriptions", workflow_message)

    # Check translations
    translations = generate_translations("en-GB")["workflow"]
    workflow_assertions = []
    for key in orchestrator.workflows.ALL_WORKFLOWS:
        if key not in translations:
            workflow_assertions.append(key)

    if workflow_assertions:
        workflow_message = "\n".join(workflow_assertions)
        raise ProcessFailure("Workflows with missing translations", workflow_message)

    return {"check_workflows_for_matching_targets_and_descriptions": True}


@step("Check that all products have at least one workflow")
def check_that_products_have_at_least_one_workflow() -> State:
    prods_without_wf = list(
        flatten(ProductTable.query.filter(not_(ProductTable.workflows.any())).with_entities(ProductTable.name))
    )
    if prods_without_wf:
        raise ProcessFailure("Found products that do not have a workflow associated with them", prods_without_wf)

    return {"check_that_products_have_at_least_one_workflow": True}


@step("Check that all products have a create, modify, terminate and validate workflow")
def check_that_products_have_create_modify_and_terminate_workflows() -> State:
    product_data = ProductTable.query.filter(ProductTable.status == "active")

    workflows_not_complete: List = []
    for product in product_data:
        workflows = {
            c.target
            for c in product.workflows
            if c.target in ["CREATE", "TERMINATE", "MODIFY", "SYSTEM"] and c.name != "modify_note"
        }
        if len(workflows) < 4:
            workflows_not_complete.append(product.name)

    # Do not raise an error but only report it in the `State` to allow exceptions.
    return {
        "products_without_at_least_create_modify_terminate_validate_workflows": workflows_not_complete,
        "check_that_products_have_create_modify_and_terminate_workflows": True,
    }


@step("Check that all active products have a modify note")
def check_that_active_products_have_a_modify_note() -> State:
    modify_workflow = WorkflowTable.query.filter(WorkflowTable.name == "modify_note").first()

    product_data = ProductTable.query.filter(ProductTable.status == "active").all()
    result = [product.name for product in product_data if modify_workflow not in product.workflows]
    if result:
        raise ProcessFailure("Found products that do not have a modify_note workflow", result)

    return {"check_that_active_products_have_a_modify_note": True}


@step("Check the DB fixed input config")
def check_db_fixed_input_config() -> State:
    fixed_input_configuration = fi_configuration()
    product_tags = products.get_tags()
    fixed_inputs = FixedInputTable.query.options(joinedload(FixedInputTable.product)).all()

    data: Dict = {"fixed_inputs": [], "by_tag": {}}
    errors: List = []

    for tag in product_tags:
        data["by_tag"][tag] = []
    for fi in fixed_inputs:
        fi_data: Dict = first_true(
            fixed_input_configuration["fixed_inputs"], {}, lambda i: i["name"] == fi.name  # noqa: B023
        )
        if not fi_data:
            errors.append(fi)

        if fi.value not in fi_data["values"]:
            errors.append(fi)

        tag_data = {one(fi) for fi in fixed_input_configuration["by_tag"][fi.product.tag]}
        tag_data_required = {one(fi) for fi in fixed_input_configuration["by_tag"][fi.product.tag] if fi[one(fi)]}

        if not tag_data:
            errors.append(fi)

        if {fi.name for fi in fi.product.fixed_inputs} - set(tag_data):
            errors.append(fi.product.name)
        if set(tag_data_required) - {fi.name for fi in fi.product.fixed_inputs}:
            errors.append(fi.product.name)

    if errors:
        raise ProcessFailure("Errors in fixed input config", errors)

    return {"check_db_fixed_input_config": True}


@step("Check subscription models")
def check_subscription_models() -> State:
    subscriptions = SubscriptionTable.query.all()
    failures: Dict[str, JSON] = {}
    for subscription in subscriptions:
        try:
            SubscriptionModel.from_subscription(subscription.subscription_id)
        except ValidationError as e:
            failures[str(subscription.subscription_id)] = e.errors()
        except Exception as e:
            failures[str(subscription.subscription_id)] = str(e)

    if failures:
        raise ProcessFailure("Found subscriptions that could not be loaded", failures)

    return {"check_subscription_models": True}


@workflow("Validate products", target=Target.SYSTEM)
def task_validate_products() -> StepList:
    return (
        init
        >> check_all_workflows_are_in_db
        >> check_workflows_for_matching_targets_and_descriptions
        >> check_that_products_have_at_least_one_workflow
        >> check_that_active_products_have_a_modify_note
        >> check_db_fixed_input_config
        >> check_that_products_have_create_modify_and_terminate_workflows
        >> check_subscription_models
        >> done
    )



from http import HTTPStatus
from uuid import uuid4

import pytest

from orchestrator.db import ProductBlockTable, ResourceTypeTable, db

RESOURCE_TYPE_ID = "f51f9542-e83f-42e5-a590-0284dd5493e4"


@pytest.fixture
def seed():
    # Delete current resource types
    ResourceTypeTable.query.delete()
    resource_types = [ResourceTypeTable(resource_type_id=RESOURCE_TYPE_ID, resource_type="some_resource_type")]
    product_block = ProductBlockTable(
        name="Ethernet", description="desc", status="active", resource_types=resource_types
    )
    db.session.add(product_block)
    db.session.commit()


def test_fetch_all_resource_types(seed, test_client):
    response = test_client.get("/api/resource_types")

    assert HTTPStatus.OK == response.status_code
    resource_types = response.json()

    assert 1 == len(resource_types)


def test_save_resource_type(seed, test_client):
    body = {"resource_type": "some"}
    response = test_client.post("/api/resource_types/", json=body)

    assert HTTPStatus.NO_CONTENT == response.status_code
    assert 2 == len(test_client.get("/api/resource_types").json())


def test_save_invalid_resource_type(seed, test_client):
    response = test_client.post("/api/resource_types/", json={})

    assert HTTPStatus.UNPROCESSABLE_ENTITY == response.status_code
    assert {
        "detail": [{"loc": ["body", "resource_type"], "msg": "field required", "type": "value_error.missing"}]
    } == response.json()


def test_update_resource_type(seed, test_client):
    body = {"resource_type_id": RESOURCE_TYPE_ID, "resource_type": "changed"}
    response = test_client.put("/api/resource_types/", json=body)

    assert HTTPStatus.NO_CONTENT == response.status_code
    resource_type = test_client.get(f"/api/resource_types/{RESOURCE_TYPE_ID}").json()

    assert "changed" == resource_type["resource_type"]


def test_delete_resource_type(seed, test_client):
    rt_id = uuid4()
    body = {"resource_type_id": rt_id, "resource_type": "some"}
    test_client.post("/api/resource_types/", json=body)
    response = test_client.delete(f"/api/resource_types/{rt_id}")
    assert HTTPStatus.NO_CONTENT == response.status_code


def test_delete_product_block_with_used_products(seed, test_client):
    response = test_client.delete(f"/api/resource_types/{RESOURCE_TYPE_ID}")
    assert HTTPStatus.BAD_REQUEST == response.status_code

    assert f"ResourceType {RESOURCE_TYPE_ID} is used in ProductBlocks: Ethernet" == response.json()["detail"]



from orchestrator.config.assignee import Assignee
from orchestrator.forms import FormPage
from orchestrator.utils.functional import orig
from orchestrator.workflow import begin, done, inputstep, step, workflow
from orchestrator.workflows.steps import unsync
from test.unit_tests.workflows import (
    WorkflowInstanceForTests,
    assert_complete,
    assert_suspended,
    resume_workflow,
    run_workflow,
)


def test_resume_workflow():
    @step("Test step 1")
    def fakestep1():
        return {"steden": ["Utrecht"], "stad": "Amsterdam"}

    @inputstep("Wait for me step 2", assignee=Assignee.SYSTEM)
    def waitforme(steden, stad):
        class WaitForm(FormPage):
            stad: str

        user_input = yield WaitForm
        meer = [*steden, stad]
        return {**user_input.dict(), "steden": meer}

    @step("Test step 3")
    def fakestep3(steden):
        meer = [*steden, "Leiden"]
        return {"steden": meer}

    @workflow("Test wf", target="Target.CREATE")
    def testwf():
        return begin >> fakestep1 >> waitforme >> fakestep3 >> done

    with WorkflowInstanceForTests(testwf, "testwf"):
        # The process and step_log from run_workflow can be used to resume it with resume_workflow
        result, process, step_log = run_workflow("testwf", {})
        assert_suspended(result)
        result, step_log = resume_workflow(process, step_log, {"stad": "Maastricht"})
        assert_complete(result)


def test_unsync(generic_subscription_1):
    result = orig(unsync)(generic_subscription_1)
    # Test is backup is available
    assert result["__old_subscriptions__"][generic_subscription_1]["description"] == "Generic Subscription One"
    assert result["__old_subscriptions__"][generic_subscription_1]["insync"] is True
    # Test if subscription will be set to insync = False:
    assert result["subscription"].insync is False



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that implements subscription related API endpoints."""
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

import structlog
from fastapi import Depends
from fastapi.param_functions import Body
from fastapi.routing import APIRouter
from oauth2_lib.fastapi import OIDCUserModel
from sqlalchemy import select
from sqlalchemy.orm import contains_eager, defer, joinedload
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response

from orchestrator.api.error_handling import raise_status
from orchestrator.api.helpers import _query_with_filters
from orchestrator.db import (
    ProcessStepTable,
    ProcessSubscriptionTable,
    ProcessTable,
    SubscriptionInstanceTable,
    SubscriptionTable,
    db,
)
from orchestrator.domain.base import SubscriptionModel
from orchestrator.schemas import SubscriptionDomainModelSchema, SubscriptionSchema, SubscriptionWorkflowListsSchema
from orchestrator.security import oidc_user
from orchestrator.services.subscriptions import (
    _generate_etag,
    build_extendend_domain_model,
    get_subscription,
    query_depends_on_subscriptions,
    query_in_use_by_subscriptions,
    subscription_workflows,
)
from orchestrator.settings import app_settings
from orchestrator.types import SubscriptionLifecycle
from orchestrator.utils.redis import from_redis

router = APIRouter()

logger = structlog.get_logger(__name__)


def _delete_subscription_tree(subscription: SubscriptionTable) -> None:
    db.session.delete(subscription)
    db.session.commit()


def _delete_process_subscriptions(process_subscriptions: List[ProcessSubscriptionTable]) -> None:
    for process_subscription in process_subscriptions:
        pid = str(process_subscription.pid)
        subscription_id = str(process_subscription.subscription_id)
        ProcessSubscriptionTable.query.filter(ProcessSubscriptionTable.pid == pid).delete()
        ProcessStepTable.query.filter(ProcessStepTable.pid == pid).delete()
        ProcessTable.query.filter(ProcessTable.pid == pid).delete()
        subscription = SubscriptionTable.query.filter(SubscriptionTable.subscription_id == subscription_id).first()
        _delete_subscription_tree(subscription)


def _filter_statuses(filter_statuses: Optional[str] = None) -> List[str]:
    """
    Check valid filter statuses.

    Args:
        filter_statuses: the filters.

    Returns:
        list of filters

    """
    if not filter_statuses:
        return []

    logger.debug("Filters to query subscription on.", filter_statuses=filter_statuses)
    statuses = filter_statuses.split(",")
    for status in statuses:
        if status not in SubscriptionLifecycle.values():
            raise HTTPException(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                detail=f"Status {status}, is not a valid `SubscriptionLifecycle`",
            )
    return statuses


@router.get("/all", response_model=List[SubscriptionSchema])
def subscriptions_all() -> List[SubscriptionTable]:
    """Return subscriptions with only a join on products."""
    return SubscriptionTable.query.all()


@router.get("/domain-model/{subscription_id}", response_model=SubscriptionDomainModelSchema)
def subscription_details_by_id_with_domain_model(
    request: Request, subscription_id: UUID, response: Response
) -> Optional[Dict[str, Any]]:
    def _build_response(model: dict, etag: str) -> Optional[Dict[str, Any]]:
        if etag == request.headers.get("If-None-Match"):
            response.status_code = HTTPStatus.NOT_MODIFIED
            return None
        response.headers["ETag"] = etag
        return model

    if cache_response := from_redis(subscription_id):
        return _build_response(*cache_response)

    try:
        subscription_model = SubscriptionModel.from_subscription(subscription_id)
        extended_model = build_extendend_domain_model(subscription_model)
        etag = _generate_etag(extended_model)
        return _build_response(extended_model, etag)
    except ValueError as e:
        if str(e) == f"Subscription with id: {subscription_id}, does not exist":
            raise_status(HTTPStatus.NOT_FOUND, f"Subscription with id: {subscription_id}, not found")
        else:
            raise_status(HTTPStatus.INTERNAL_SERVER_ERROR, str(e))


@router.delete("/{subscription_id}", response_model=None)
def delete_subscription(subscription_id: UUID) -> None:
    all_process_subscriptions = ProcessSubscriptionTable.query.filter_by(subscription_id=subscription_id).all()
    if len(all_process_subscriptions) > 0:
        _delete_process_subscriptions(all_process_subscriptions)
        return None
    else:
        subscription = SubscriptionTable.query.filter(SubscriptionTable.subscription_id == subscription_id).first()
        if not subscription:
            raise_status(HTTPStatus.NOT_FOUND)

        _delete_subscription_tree(subscription)
        return None


@router.get("/parent_subscriptions/{subscription_id}", response_model=List[SubscriptionSchema], deprecated=True)
@router.get("/in_use_by/{subscription_id}", response_model=List[SubscriptionSchema])
def in_use_by_subscriptions(
    subscription_id: UUID, filter_statuses: List[str] = Depends(_filter_statuses)
) -> List[SubscriptionTable]:
    """
    Retrieve subscriptions that are in use by this subscription.

    Args:
        subscription_id: Subscription to query
        filter_statuses: List of filters

    Returns:
        list of subscriptions

    """
    return query_in_use_by_subscriptions(subscription_id, filter_statuses).all()


@router.post("/subscriptions_for_in_used_by_ids", response_model=Dict[UUID, SubscriptionSchema])
def subscriptions_by_in_used_by_ids(data: List[UUID] = Body(...)) -> Dict[UUID, SubscriptionSchema]:
    rows = db.session.execute(
        select(SubscriptionInstanceTable)
        .join(SubscriptionTable)
        .filter(SubscriptionInstanceTable.subscription_instance_id.in_(data))
    ).all()
    result = {row[0].subscription_instance_id: row[0].subscription for row in rows}
    if len(rows) != len(data):
        logger.warning(
            "Not all subscription_instance_id's could be resolved.",
            unresolved_ids=list(set(data) - set(result.keys())),
        )
    return result


@router.get("/child_subscriptions/{subscription_id}", response_model=List[SubscriptionSchema], deprecated=True)
@router.get("/depends_on/{subscription_id}", response_model=List[SubscriptionSchema])
def depends_on_subscriptions(
    subscription_id: UUID,
    filter_statuses: List[str] = Depends(_filter_statuses),
) -> List[SubscriptionTable]:
    """
    Retrieve dependant subscriptions.

    Args:
        subscription_id: The subscription id
        filter_statuses: the status of dependant subscriptions

    Returns:
        List of dependant subscriptions.

    """
    return query_depends_on_subscriptions(subscription_id, filter_statuses).all()


@router.get("/", response_model=List[SubscriptionSchema])
def subscriptions_filterable(
    response: Response, range: Optional[str] = None, sort: Optional[str] = None, filter: Optional[str] = None
) -> List[SubscriptionTable]:
    """
    Get subscriptions filtered.

    Args:
        response: Fastapi Response object
        range: Range
        sort: Sort
        filter: Filter

    Returns:
        List of subscriptions

    """
    _range: Union[List[int], None] = list(map(int, range.split(","))) if range else None
    _sort: Union[List[str], None] = sort.split(",") if sort else None
    _filter: Union[List[str], None] = filter.split(",") if filter else None
    logger.info("subscriptions_filterable() called", range=_range, sort=_sort, filter=_filter)
    query = SubscriptionTable.query.join(SubscriptionTable.product).options(
        contains_eager(SubscriptionTable.product), defer("product_id")
    )
    query_result = _query_with_filters(response, query, _range, _sort, _filter)
    return query_result


@router.get(
    "/workflows/{subscription_id}", response_model=SubscriptionWorkflowListsSchema, response_model_exclude_none=True
)
def subscription_workflows_by_id(subscription_id: UUID) -> Dict[str, List[Dict[str, Union[List[Any], str]]]]:
    subscription = SubscriptionTable.query.options(joinedload("product"), joinedload("product.workflows")).get(
        subscription_id
    )
    if not subscription:
        raise_status(HTTPStatus.NOT_FOUND)

    return subscription_workflows(subscription)


@router.get("/instance/other_subscriptions/{subscription_instance_id}", response_model=List[UUID])
def subscription_instance_in_use_by(
    subscription_instance_id: UUID, filter_statuses: List[str] = Depends(_filter_statuses)
) -> List[UUID]:
    subscription_instance: SubscriptionInstanceTable = SubscriptionInstanceTable.query.get(subscription_instance_id)

    if not subscription_instance:
        raise_status(HTTPStatus.NOT_FOUND)

    in_use_by_instances = subscription_instance.in_use_by
    if filter_statuses:
        in_use_by_instances = [sub for sub in in_use_by_instances if sub.subscription.status in filter_statuses]

    return list(
        filter(
            lambda sub_id: sub_id != subscription_instance.subscription_id,
            {sub.subscription_id for sub in in_use_by_instances},
        )
    )


@router.put("/{subscription_id}/set_in_sync", response_model=None, status_code=HTTPStatus.OK)
def subscription_set_in_sync(subscription_id: UUID, current_user: Optional[OIDCUserModel] = Depends(oidc_user)) -> None:
    def failed_processes() -> List[str]:
        if app_settings.DISABLE_INSYNC_CHECK:
            return []
        _failed_processes = (
            ProcessSubscriptionTable.query.join(ProcessTable)
            .filter(ProcessSubscriptionTable.subscription_id == subscription_id)
            .filter(~ProcessTable.is_task)
            .filter(ProcessTable.last_status != "completed")
            .filter(ProcessTable.last_status != "aborted")
            .all()
        )
        return [str(p.pid) for p in _failed_processes]

    try:
        subscription = get_subscription(subscription_id, for_update=True)
        if not subscription.insync:
            logger.info(
                "Subscription not in sync, trying to change..", subscription_id=subscription_id, user=current_user
            )
            failed_processes = failed_processes()  # type: ignore
            if not failed_processes:
                subscription.insync = True  # type: ignore
                db.session.commit()
                logger.info("Subscription set in sync", user=current_user)
            else:
                raise_status(
                    HTTPStatus.UNPROCESSABLE_ENTITY,
                    f"Subscription {subscription_id} has still failed processes with id's: {failed_processes}",
                )
        else:
            logger.info("Subscription already in sync")
    except ValueError as e:
        raise_status(HTTPStatus.NOT_FOUND, str(e))



from http import HTTPStatus
from os import getenv

import pytest
from redis import Redis

from orchestrator.db import WorkflowTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.settings import app_settings
from orchestrator.targets import Target
from orchestrator.utils.functional import orig
from orchestrator.workflows.steps import cache_domain_models

PRODUCT_ID = "fb28e465-87fd-4d23-9c75-ed036529e416"


def test_workflows(test_client):
    response = test_client.get("/api/workflows")

    assert HTTPStatus.OK == response.status_code
    workflows = response.json()

    assert len(workflows) == WorkflowTable.query.count()
    for workflow in workflows:
        assert workflow["name"] is not None
        assert workflow["target"] is not None


def test_workflows_by_target(test_client):
    for target, num_wfs in [
        (Target.CREATE, WorkflowTable.query.filter(WorkflowTable.target == Target.CREATE).count()),
        (Target.TERMINATE, WorkflowTable.query.filter(WorkflowTable.target == Target.TERMINATE).count()),
        (Target.MODIFY, WorkflowTable.query.filter(WorkflowTable.target == Target.MODIFY).count()),
    ]:
        response = test_client.get(f"/api/workflows?target={target}")
        workflows = response.json()
        assert len(workflows) == num_wfs
        for wf in workflows:
            assert target == wf["target"]


def test_get_all_with_product_tags(test_client):
    response = test_client.get("/api/workflows/with_product_tags")

    assert HTTPStatus.OK == response.status_code
    assert len(response.json()) == WorkflowTable.query.count()


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", 0) == 0, reason="AIOCACHE must be enabled for this test to do anything"
)
def test_push_subscriptions_to_cache_step(generic_subscription_1):
    push_subscription = orig(cache_domain_models)
    push_subscription("Aribtrary_name", SubscriptionModel.from_subscription(generic_subscription_1))

    cache = Redis(host=app_settings.CACHE_HOST, port=app_settings.CACHE_PORT)
    assert cache.get(f"domain:{generic_subscription_1}") is None

    app_settings.CACHE_DOMAIN_MODELS = True
    push_subscription("Aribtrary_name", SubscriptionModel.from_subscription(generic_subscription_1))
    result = cache.get(f"domain:{generic_subscription_1}")
    assert result is not None
    app_settings.CACHE_DOMAIN_MODELS = False



import functools
from typing import Optional

import pytest

from orchestrator.utils.functional import (
    as_t,
    expand_ranges,
    first_available_or_next,
    ireplace,
    join_cs,
    orig,
    to_ranges,
)


def test_join_cs():
    assert join_cs("") == ""
    assert join_cs([]) == ""
    assert join_cs(()) == ""

    assert join_cs("", []) == ""
    assert join_cs([], "") == ""
    assert join_cs("", ()) == ""
    assert join_cs((), "") == ""

    assert join_cs("a") == "a"
    assert join_cs(["a"]) == "a"
    assert join_cs(("a",)) == "a"

    assert join_cs("a", "b") == "a,b"
    assert join_cs(["a"], ["b"]) == "a,b"
    assert join_cs(["a"], ("b",)) == "a,b"

    assert join_cs("a,,b") == "a,b"

    assert join_cs("a,b") == "a,b"
    assert join_cs(["a", "b"]) == "a,b"
    assert join_cs(("a", "b")) == "a,b"

    assert join_cs("a,b", ["c", "d"]) == "a,b,c,d"
    assert join_cs(["a", "b"], "c,d") == "a,b,c,d"
    assert join_cs(("a", "b"), "c,d") == "a,b,c,d"

    assert join_cs("a,b", ["c", "d"], ("e", "f")) == "a,b,c,d,e,f"

    with pytest.raises(TypeError):
        join_cs(1, 2)

    with pytest.raises(TypeError):
        join_cs([1])


def test_first_available_or_next():
    assert first_available_or_next([0, 1, 3]) == 2
    assert first_available_or_next([0, 1, 2, 3]) == 4
    assert first_available_or_next([1, 2, 3]) == 0
    assert first_available_or_next([]) == 0

    assert first_available_or_next([0, 1, 3], start=11) == 11
    assert first_available_or_next([0, 1, 3], start=4) == 4
    assert first_available_or_next([], 22) == 22
    assert first_available_or_next([1, 100, 101], 33) == 33
    assert first_available_or_next([11, 22, 33, 44, 55], 33) == 34


def test_expand_ranges():
    assert expand_ranges([[1], [2], [10, 12]]) == [1, 2, 10, 11]
    assert expand_ranges([[1], [2], [10, 12]], inclusive=True) == [1, 2, 10, 11, 12]
    assert expand_ranges([[1], [2], [10, 12]], inclusive=True) == [1, 2, 10, 11, 12]
    assert expand_ranges([]) == []

    # sorted
    assert expand_ranges([[100], [1, 4]], inclusive=True) == [1, 2, 3, 4, 100]

    # deduplicated
    assert expand_ranges([[1, 5], [3, 5]], inclusive=True) == [1, 2, 3, 4, 5]

    with pytest.raises(ValueError):
        expand_ranges([[]])

    with pytest.raises(ValueError):
        expand_ranges([[2, 100, 3]])


def test_as_t():
    # Don't know how to check type annotations at runtime yet. Hence test only basic functionality of `as_t` and not
    # the casting of Optional[T] to just T
    x: Optional[int] = 7
    y: int = as_t(x)
    assert y == 7
    with pytest.raises(ValueError):
        as_t(None)


def test_ireplace():
    assert list(ireplace(["1-10", "", "22"], "", "0")) == ["1-10", "0", "22"]

    # Values are tested in their entirety, hence "10" won't be replaced in the value "1-10"
    assert list(ireplace(["1-10", "", "22"], "10", "999")) == ["1-10", "", "22"]


def test_to_ranges():
    assert list(to_ranges([1, 2, 3])) == [range(1, 4)]
    assert list(to_ranges([])) == []
    assert list(to_ranges([0])) == [range(0, 1)]
    assert list(to_ranges([1, 2, 3, 7, 8, 9, 100, 200, 201, 202])) == [
        range(1, 4),
        range(7, 10),
        range(100, 101),
        range(200, 203),
    ]


def test_orig():
    def func():
        pass

    @functools.wraps(func)
    def wrapper():
        return func()

    @functools.wraps(wrapper)
    def super_wrapper():
        return wrapper()

    assert orig(wrapper) == func
    assert orig(super_wrapper) == func



from dataclasses import dataclass
from typing import Optional


@dataclass
class ImsPort:
    node_id: int
    status: int
    iface_type: str
    node: str


@dataclass
class ImsNode:
    location: str
    status: str
    id: Optional[int]


IP_PREFIX_NAME = "IP Prefix"

IP_PREFIX_PRODUCT_ID = "2b2125f2-a074-4e44-8d4b-edc677381d46"
IPBGP_PRODUCT_ID = "876125f2-a074-4e44-8d4b-edc677381d46"
IPS_PRODUCT_ID = "456125f2-a074-4e44-8d4b-edc677381d46"
SP_PRODUCT_ID = "6911deed-e2b5-4038-9fca-6cb7bc6c07d3"
AGGSP_PRODUCT_ID = "1367deed-e2b5-4038-9fca-6cb7bc6c07d3"

IP_PREFIX_SUB_ID = "e89776be-16c3-4bee-af98-8e73bf6492a7"
IPS_SUB_ID = "abc776be-16c3-4bee-af98-8e73bf6492a7"
IPBGP_SUB_ID = "123776be-16c3-4bee-af98-8e73bf6492a7"
MISSING_SUB_ID = "00e651c6-e16a-4ff6-8f21-af78ff199faf"
TAGGED_SUB_ID = "b2791099-86a3-4754-b70b-529bf1126246"
UNTAGGED_SUB_ID = "de6bcd9a-e003-49b0-a5e0-0bf43f747d80"
AGGSP_SUB_ID = "1858cd9a-e003-49b0-a5e0-0bf43f747d80"

CUSTOMER_ID1 = "1f2c4fae-77ab-4edf-b67a-f99ad3e49ad1"


PRODUCT_TO_TAG = {
    IP_PREFIX_PRODUCT_ID: "IP_PREFIX",
    IPBGP_PRODUCT_ID: "IPBGP",
    IPS_PRODUCT_ID: "IPS",
    SP_PRODUCT_ID: "SP",
    AGGSP_PRODUCT_ID: "AGGSP",
}

SUB_TO_PRODUCT = {
    IP_PREFIX_SUB_ID: IP_PREFIX_PRODUCT_ID,
    IPS_SUB_ID: IPS_PRODUCT_ID,
    IPBGP_SUB_ID: IPBGP_PRODUCT_ID,
    MISSING_SUB_ID: IP_PREFIX_PRODUCT_ID,
    TAGGED_SUB_ID: SP_PRODUCT_ID,
    UNTAGGED_SUB_ID: SP_PRODUCT_ID,
    AGGSP_SUB_ID: AGGSP_PRODUCT_ID,
}



from orchestrator.api.helpers import product_block_paths


def test_product_block_paths(sub_list_union_overlap_subscription_1):
    paths = product_block_paths(sub_list_union_overlap_subscription_1)
    assert paths == [
        "product",
        "test_block.sub_block",
        "test_block.sub_block_2",
        "test_block.sub_block_list.0",
        "test_block",
        "list_union_blocks.0",
        "list_union_blocks.1",
    ]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from shutil import copyfile
from typing import List, Optional

import jinja2
import typer
from alembic import command
from alembic.config import Config
from structlog import get_logger

import orchestrator

logger = get_logger(__name__)

app: typer.Typer = typer.Typer()

orchestrator_module_location = os.path.dirname(orchestrator.__file__)
migration_dir = "migrations"

loader = jinja2.FileSystemLoader(os.path.join(orchestrator_module_location, f"{migration_dir}/templates"))
jinja_env = jinja2.Environment(
    loader=loader, autoescape=True, lstrip_blocks=True, trim_blocks=True, undefined=jinja2.StrictUndefined
)


def alembic_cfg() -> Config:
    cfg = Config("alembic.ini")
    version_locations = cfg.get_main_option("version_locations")
    cfg.set_main_option(
        "version_locations", f"{version_locations} {orchestrator_module_location}/{migration_dir}/versions/schema"
    )
    logger.info("Version Locations", locations=cfg.get_main_option("version_locations"))
    return cfg


@app.command(
    help="Initialize an empty migrations environment. This command will throw an exception when it detects conflicting files and directories."
)
def init() -> None:
    """
    Initialise the migrations directory.

    This command will initialize a migration directory for the orchestrator core application and setup a correct
    migration environment.

    Returns:
        None

    """

    if os.access(migration_dir, os.F_OK) and os.listdir(migration_dir):
        raise OSError(f"Directory {migration_dir} already exists and is not empty")

    logger.info("Creating directory", directory=os.path.abspath(migration_dir))
    os.makedirs(migration_dir)
    versions = os.path.join(migration_dir, "versions")
    logger.info("Creating directory", directory=os.path.abspath(versions))
    os.makedirs(versions)
    versions_schema = os.path.join(migration_dir, "versions/schema")
    logger.info("Creating directory", directory=os.path.abspath(versions_schema))
    os.makedirs(versions_schema)

    source_env_py = os.path.join(orchestrator_module_location, f"{migration_dir}/templates/env.py.j2")
    env_py = os.path.join(migration_dir, "env.py")
    logger.info("Creating file", file=os.path.abspath(env_py))
    copyfile(source_env_py, env_py)

    source_script_py_mako = os.path.join(orchestrator_module_location, f"{migration_dir}/script.py.mako")
    script_py_mako = os.path.join(migration_dir, "script.py.mako")
    logger.info("Creating file", file=os.path.abspath(script_py_mako))
    copyfile(source_script_py_mako, script_py_mako)

    source_helpers_py = os.path.join(orchestrator_module_location, f"{migration_dir}/templates/helpers.py.j2")
    helpers_py = os.path.join(migration_dir, "helpers.py")
    logger.info("Creating file", file=os.path.abspath(helpers_py))
    copyfile(source_helpers_py, helpers_py)

    template = jinja_env.get_template("alembic.ini.j2")

    if not os.access(os.path.join(os.getcwd(), "alembic.ini"), os.F_OK):
        logger.info("Creating file", file=os.path.join(os.getcwd(), "alembic.ini"))
        with open(os.path.join(os.getcwd(), "alembic.ini"), "w") as alembic_ini:
            alembic_ini.write(template.render(migrations_dir=migration_dir))
    else:
        logger.info("Skipping Alembic.ini file. It already exists")


@app.command(help="Get the database heads")
def heads() -> None:
    command.heads(alembic_cfg())


@app.command(help="Merge database revisions.")
def merge(
    revisions: Optional[List[str]] = typer.Argument(
        None, help="Add the revision you would like to merge to this command."
    ),
    message: str = typer.Option(None, "--message", "-m", help="The revision message"),
) -> None:
    """
    Merge database revisions.

    Args:
        revisions: List of revisions to merge
        message: Optional message for the revision.

    Returns:
        None

    """
    command.merge(alembic_cfg(), revisions, message=message)


@app.command()
def upgrade(revision: Optional[str] = typer.Argument(None, help="Rev id to upgrade to")) -> None:
    """
    Upgrade the database.

    Args:
        revision: Optional argument to indicate where to upgrade to.

    Returns:
        None

    """
    command.upgrade(alembic_cfg(), revision)


@app.command()
def downgrade(revision: Optional[str] = typer.Argument(None, help="Rev id to upgrade to")) -> None:
    """
    Downgrade the database.

    Args:
        revision: Optional argument to indicate where to downgrade to.

    Returns:
        None

    """
    command.downgrade(alembic_cfg(), revision)


@app.command()
def revision(
    message: str = typer.Option(None, "--message", "-m", help="The revision message"),
    version_path: str = typer.Option(None, "--version-path", help="Specify specific path from config for version file"),
    autogenerate: bool = typer.Option(False, help="Detect schema changes and add migrations"),
    head: str = typer.Option(None, help="Determine the head the head you need to add your migration to."),
) -> None:
    """
    Create a new revision file.

    Args:
        message: The revision message
        version_path: Specify specific path from config for version file
        autogenerate: Whether to detect schema changes.
        head: To which head the migration applies

    Returns:
        None

    """
    command.revision(alembic_cfg(), message, version_path=version_path, autogenerate=autogenerate, head=head)


@app.command()
def history(
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose output"),
    indicate_current: bool = typer.Option(False, "--current", "-c", help="Indicate current revision"),
) -> None:
    """
    List changeset scripts in chronological order.

    Args:
        verbose: Verbose output
        indicate_current: Indicate current revision

    Returns:
        None

    """
    command.history(alembic_cfg(), verbose=verbose, indicate_current=indicate_current)
